/*
** NOE 2.1 Linux kernel driver
** ether part
**
** Copyright (C) Ubiquitous Corp. 2001-2011. All rights reserved.
**
** $Revision: #1 $	$Date: 2012/08/03 $
**
*/
//  original driver's comment.
/*
 * network driver -
 *
 *  Copyright (c) 2002 Petko Manolov (petkan@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * for NOE: use from [2007-10-24]
 */
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/mutex.h>
#include <linux/sockios.h>
#include <linux/stat.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/interrupt.h>
#include <linux/version.h>
#include <linux/ethtool.h>
#include <linux/wireless.h>
#include <linux/rtnetlink.h>
#include <linux/kthread.h>
#include <net/iw_handler.h>

#include "noe_std.h"
#include "noe_dev.h"
#include "noe_driver_if.h"
#include "noe_drv_config.h"
#include "wlan/wlandef.h"

/* added by sony */
#include "net_debug.h"
#include "noe_command_debug.h"

#define	ETHER_REQT_READ		0xc0
#define	ETHER_REQT_WRITE	0x40
#define	ETHER_REQ_GET_REGS	0x05
#define	ETHER_REQ_SET_REGS	0x05

typedef struct {
	NOE_HEADER	h;
	NOE_UINT32	link_status;
						// payload must be multiply 4
} NOE_LINK_STATUS;

void noe_rec_data(struct urb *urb, noe_drv_t *noe_drv, noe_rcv_usb_data *data); /* modified by sony */
extern const char driver_name[];
extern const char driver_version[];

static void ether_get_drvinfo(struct net_device *netdev, struct ethtool_drvinfo *info);
static int ether_get_settings(struct net_device *netdev, struct ethtool_cmd *ecmd);
static int ether_open(struct net_device *netdev);
static int ether_close(struct net_device *netdev);
static void ether_tx_timeout(struct net_device *netdev);
int ether_hard_start_xmit(struct sk_buff *skb, struct net_device *netdev);
static struct net_device_stats *ether_get_stats(struct net_device *netdev);

static struct ethtool_ops ops = {
	.get_drvinfo = ether_get_drvinfo,
	.get_settings = ether_get_settings,
	.get_link = ethtool_op_get_link
};

/***************************************************/
/*            new defines by SONY                  */
/***************************************************/
/* added new structure by sony */
typedef struct {
    noe_drv_t        *noe_drv;
  //    void      *payload;
    noe_rcv_usb_data *data;
} noe_rcv_usb_param;

/***************************************************/
/*           new functions by SONY                 */
/***************************************************/
static void _validate_noe_system( noe_drv_t *noe_drv, int status, void *buffer );

static struct task_struct* _create_rx_fixup_thread( noe_drv_t *noe_drv );
static void _destroy_rx_fixup_thread( noe_drv_t *noe_drv );
static void _request_rx_fixup( noe_drv_t *noe_drv );
static int _rx_fixup_thread_main(void* data);
static inline noe_rcv_usb_data* _allocate_rx_buf( noe_drv_t *noe_drv );
static inline void _free_rx_buf( noe_drv_t *noe_drv, noe_rcv_usb_data *buf );

static bool _is_hdrl_rec_pause_condition( noe_drv_t *noe_drv );
static void _check_rec_pause( unsigned long data );
static void _clear_check_hdrl_rec_pause_timer( noe_drv_t *noe_drv );

#ifdef CONFIG_NOE_DEBUG_MEMORY
static void _check_rcv_buffer( noe_drv_t *noe_drv );
#endif

bool is_init_usb_status( void ); /* usb_drv.c */

/********************** debug control by sony ******************/
#define NOE_DEBUG_ERROR
#define NOE_DEBUG_WARNING
#define NOE_DEBUG_INFO
#undef  NOE_DEBUG_DP
#undef  NOE_DEBUG_PASS_THROUGH
#undef  NOE_DEBUG_HDRL_SVR
#undef  NOE_DEBUG_DLNA_SVR
#undef  NOE_DEBUG_TRACE_BULK_TRANSFER
#include "noe_debug.h"

/*
**
**	device related part of the code
**
*/
static int
get_registers(noe_drv_t *noe_drv, u16 value, u16 index, u16 size, void *data)
{
	return(usb_control_msg(noe_drv->udev, usb_rcvctrlpipe(noe_drv->udev, 0),
				       ETHER_REQ_GET_REGS, ETHER_REQT_READ,
			               value, index, data, size, 5*HZ));
}

#if defined(CONFIG_PREEMPT)
static int
set_registers(noe_drv_t *noe_drv, u16 value, u16 index, u16 size, void *data)
{
	return(usb_control_msg(noe_drv->udev, usb_sndctrlpipe(noe_drv->udev, 0),
					ETHER_REQ_SET_REGS, ETHER_REQT_WRITE,
					value, index, data, size, 5*HZ));
}
#endif

static void
send_noe_response(noe_drv_t *noe_drv, NOE_INT32 nr, NOE_UINT32 ssid)
{
	NOE_HEADER		h;

	h.SubSysID = ssid;
	h.HostID = (NOE_UINT32)&h;
	h.cmd = nr;
	h.len = 0;
	h.ctrl = 0;

	NOEki_send_noe_header_payload(noe_drv, &h, NULL);
}

static void
send_noe_response0(noe_drv_t *noe_drv, NOE_INT32 nr, NOE_UINT32 ssid, NOE_UINT32 r0)
{
	NOE_HEADER		h;

	h.SubSysID = ssid;
	h.HostID = (NOE_UINT32)&h;
	h.cmd = nr;
	h.len = 0;
	h.ctrl = 0;
	h.r0 = host2sub32(r0);

	NOEki_send_noe_header_payload(noe_drv, &h, NULL);
}

static inline void
set_ethernet_addr(noe_drv_t *noe_drv, struct net_device *netdev, int index)
{
	u8 node_id[6];

	//get mac adress from NIC
	get_registers(noe_drv, get_Mac_Address, index, sizeof(node_id), node_id);

	memcpy(netdev->dev_addr, node_id, sizeof(node_id));
	NOE_DP("set_ethernet_addr[%02x:%02x:%02x:%02x:%02x:%02x]\n", node_id[0],node_id[1],node_id[2],node_id[3],node_id[4],node_id[5]);
}

static inline unsigned char
get_num_nif(noe_drv_t *noe_drv)
{
	unsigned char nif;

	get_registers(noe_drv, get_Num_NIF, 0, sizeof(nif), &nif);

	return(nif);
}

static void
noe_set_carrier(noe_drv_t *noe_drv, NOE_UINT32 status)
{
	int	i;
	net_device_private_t	*noe_priv;

	NOE_DP("noe_set_carrier:%d\n",status);

	for(i = 0;i < MAX_NIF;i++) {
          /* modified by sony */
                struct net_device *netdev = netdevinfo2netdev(noe_drv, i, 0);
		if(netdev) {
			noe_priv = noe_netdev_priv(netdev);
			if(atomic_read(&noe_priv->opened)) {
				if(status & (1 << i))
					netif_carrier_on(netdev);
				else
					netif_carrier_off(netdev);
			}
		}
	}
}

static void
set_carrier(struct net_device *netdev)
{
	u8	tmp;

	//get carrier status from NIC
	get_registers(net2noe_drv(netdev), get_Link_Status, noe_netdev_priv(netdev)->nif_id, 1, &tmp);

	NOE_DP("set_carrier:%02x\n",tmp);

	if(tmp == LINK_ON)
		netif_carrier_on(netdev);
	else
		netif_carrier_off(netdev);
}

static void 
noept_set_ether_blocks(noe_drv_t *noe_drv, void *rcv_buffer)
{
	char		*p;
	NOE_UINT32	ip_summed;
	NOE_UINT32	i;
	NOE_UINT32	len;
	NOE_UINT32	num_of_ether;
	NOE_HEADER	*h;
	NOE_NIFID	nif_id;
	struct sk_buff	*skb;
	struct net_device *netdev;

	p = rcv_buffer;
	h = (NOE_HEADER *)p;
	num_of_ether = sub2host32(h->r0);
	p += sizeof(NOE_HEADER);

	for(i = 0;i < num_of_ether;i++) {
		ip_summed = sub2host32(*(NOE_UINT32 *)p);
		len = ip_summed & ~(NOE_NETWORK_PACKET_OFFLOAD_CHECKSUM|NOE_NETWORK_NIC_NUMBER);
		nif_id = (ip_summed & NOE_NETWORK_NIC_NUMBER) >> NOE_NETWORK_NIC_SHIFT;
		netdev = netdevinfo2netdev( noe_drv, NOE_NIFID2NO(nif_id), NOE_NIFID2SUBNO(nif_id) );

                if( len <= 0 || len >= NFE_MULTI_PACKET_FRAME_SIZE ) continue; /* modified by sony */

		p += sizeof(NOE_UINT32);
		if(netdev && netif_device_present(netdev)) {
			skb = dev_alloc_skb(len + 2); /* modified by sony len+4->len+2 */
			if(skb == NULL) {
				NOE_ERROR("noe:skb alloc err!\n");
				return;
			}
                        skb_reserve(skb, 2); /* modified by sony for alignment */

			memcpy(skb_put(skb, len), p, len);
		
			skb->dev = netdev;
			skb->protocol = eth_type_trans(skb, netdev);
			if(ip_summed & NOE_NETWORK_PACKET_OFFLOAD_CHECKSUM){
				skb->ip_summed = CHECKSUM_UNNECESSARY;
                                NOE_PASS_DP("CheckSum Unnecessary\n");
                        }
			else{
				skb->ip_summed = CHECKSUM_NONE;
                                NOE_PASS_DP("CheckSum Necessary\n");
                        }
                        DUMP_SKB(nif_id, i+1, num_of_ether, skb,len); /* we can dump from Ethernet header at this point */
			netif_rx(skb);
			noe_netdev_priv(netdev)->stats.rx_packets++;
			noe_netdev_priv(netdev)->stats.rx_bytes += len;
		}
		p += NOE_ETHER_BLOCK_PAYLOAD;

                if( test_bit(DEF_NFE_DEBUG_NETWORK_MONITOR, &noe_drv->debug_status) ){
                    noe_drv->debug_pass_thru_rcv_len += len;
                }

#ifdef NOE_SACK_CHECK_MODE
                _debug_sack_packet_count(len);
#endif

	}
}

static void
noept_wireless_send_event(struct work_struct *work)
{
	net_device_private_t	*noe_priv;

	noe_priv = container_of(work, net_device_private_t, wqWirelessSendEvent);

        wireless_send_event(noe_priv->netdev, noe_priv->wreq_cmd, &noe_priv->wreq, noe_priv->wreq_buf);

	send_noe_response(noe_priv->noe_drv, NOE_ACK, noe_priv->wreq_ssid);
}

static void
noept_netif_update_flags(struct work_struct *work)
{
	net_device_private_t	*noe_priv;

	noe_priv = container_of(work, net_device_private_t, wqNetifUpdateFlags);

	noe_priv->netdev->flags = (noe_priv->netdev->flags & noe_priv->netif_update_flags_and) | noe_priv->netif_update_flags_or;

	send_noe_response0(noe_priv->noe_drv, NOE_ACK, noe_priv->netif_update_flags_ssid, noe_priv->netdev->flags);
}

static void
noe_update_netdev_flags(struct work_struct *work)
{
	NOE_HEADER	h;
	net_device_private_t	*noe_priv;
	noe_drv_t	*noe_drv;

	noe_priv = container_of(work, net_device_private_t, wqUpdateNetdevFlags);
	noe_drv = noe_priv->noe_drv;

	h.SubSysID = NOE_SID_UNBIND;
	h.HostID = (NOE_UINT32)&h;			// stack frame is uniq!!
	h.len = 0;
	h.cmd = NOE_BDEX_NETIF_UPDATE_FLAGS_H2S;
	h.ctrl = noe_priv->nif_id;
	h.r0 = NOEhi_htos32(noe_priv->dev_ssid);
	h.r1 = NOEhi_htos32(noe_priv->netdev->flags);

	NOEki_send_noe_header_payload(noe_drv, &h, NULL);
}

static void
noe_tx_timeout(struct work_struct *work)
{
	NOE_HEADER	h;
	net_device_private_t	*noe_priv;
	noe_drv_t	*noe_drv;

	noe_priv = container_of(work, net_device_private_t, wqTxTimeout);
	noe_drv = noe_priv->noe_drv;

	h.SubSysID = NOE_SID_UNBIND;
	h.HostID = (NOE_UINT32)&h;			// stack frame is uniq!!
	h.len = 0;
	h.cmd = NOE_BDEX_WIFI_TX_TIMEOUT;
	h.ctrl = noe_priv->nif_id;
	h.r0 = NOEhi_htos32(noe_priv->dev_ssid);
	h.r1 = NOEhi_htos32(noe_priv->netdev->flags);

	NOEki_send_noe_header_payload(noe_drv, &h, NULL);

	NOEki_wait_normal_packet(noe_drv, &h, NULL);
}

	//
	// device change event from subsystem is serialized by subsystem.
	//
static void
noept_device_changed(struct work_struct *work)
{
	NOE_DEVICE_CHANGED	*dev_changed;
	NOE_NIFID	nif_id;
	int		nif_no;
	int		nif_subno;
	NOE_UINT32	dev_ssid;
	noe_drv_t	*noe_drv;
	NOE_UINT32	nr = NOE_ACK;
	NOE_UINT8	subcmd;
	net_device_private_t *noe_priv;
	struct net_device *netdev;

	noe_drv = container_of(work, noe_drv_t, wqDevChanged);

	dev_changed = &noe_drv->noe_dc;
	nif_id = dev_changed->nif_id;
	nif_no = NOE_NIFID2NO(nif_id);
	nif_subno = NOE_NIFID2SUBNO(nif_id);
	dev_ssid = sub2host32(dev_changed->dev_ssid);

NOE_DP("nif_no=%d:%d subcmd=%x mac:%02x:%02x:%02x:%02x:%02x:%02x\n", nif_no, nif_subno, dev_changed->subcmd,
dev_changed->mac[0],
dev_changed->mac[1],
dev_changed->mac[2],
dev_changed->mac[3],
dev_changed->mac[4],
dev_changed->mac[5]);

	subcmd = dev_changed->subcmd & NOE_M_DC_CHANGED;
	if(subcmd == NOE_F_DCC_REMOVED) {
							/* this double parentheses is to suppress gcc's warning */
                if((netdev = netdevinfo2netdev(noe_drv, nif_no, nif_subno))) {
			NOE_DP("%s: %s will be deleted\n", __func__, netdev->name);
			noe_priv = netdev_priv(netdev);
			if(noe_priv->wreq_buf) {
				kfree(noe_priv->wreq_buf);
				noe_priv->wreq_buf = NULL;	/*for safe*/
			}
                        /* modified by sony */
                        if(test_bit( VAP_STATUS_DELETE_PROCESSING_BY_IOCTL, &noe_drv->vap_status )){
                            clear_bit( VAP_STATUS_DELETE_PROCESSING_BY_IOCTL, &noe_drv->vap_status );
                            unregister_netdevice(netdev);
                            rtnl_unlock(); /* flush netlink todo in kernel */
                            rtnl_lock();
                        }
                        else{
                            unregister_netdev(netdev);
                        }
                        free_netdev(netdev);
                        unregister_netdevinfo( noe_drv, nif_no, nif_subno);
		}
		else {
			NOE_WARNING("null entry is called\n");
		}
	}
	else {
                if(!is_registed_netdevinfo( noe_drv, nif_no, nif_subno )) {

                        bool rtnl_lock_already = false; /* added by sony */

			if(dev_changed->ifname[0])
				netdev = alloc_netdev(sizeof(*noe_priv), dev_changed->ifname, ether_setup);
			else
				netdev = alloc_etherdev(sizeof(*noe_priv));
			if(netdev == NULL) {
				NOE_ERROR("Out of memory\n");
				nr = NOE_NAK_NO_MEMORY;
				goto _error;
			}

			noe_priv = netdev_priv(netdev);
			memset(noe_priv, 0, sizeof(*noe_priv));
			noe_priv->noe_drv = noe_drv;
			noe_priv->nif_id = nif_id;
			noe_priv->dev_ssid = dev_ssid;
			noe_priv->netdev = netdev;
			atomic_set(&noe_priv->opened, 0);

			switch(subcmd) {
			case NOE_F_DCC_ADDED_ETH:
                                NOE_DP(" Add Ethernet Interface:%s\n", dev_changed->ifname);
				SET_MODULE_OWNER(netdev);
				netdev->open = ether_open;
				netdev->stop = ether_close;
				netdev->watchdog_timeo = ETHER_TX_TIMEOUT;
				netdev->tx_timeout = ether_tx_timeout;
				netdev->hard_start_xmit = ether_hard_start_xmit;
				netdev->get_stats = ether_get_stats;
				netdev->mtu = ETHER_MTU;
				break;
			case NOE_F_DCC_ADDED_WIFI:
                                NOE_DP(" Add WiFi Interface:%s\n", dev_changed->ifname);
				nr = initialize_wlan(netdev);
				if(nr != NOE_ACK)
				  goto _error;
				break;
			case NOE_F_DCC_ADDED_VAP:
                                NOE_DP(" Add VAP Interface:%s\n", dev_changed->ifname);
				nr = initialize_vap(netdev);
				if(nr != NOE_ACK)
				  goto _error;
                                rtnl_lock_already = true;
				break;
			default:
				NOE_ERROR("unsupported, subcmd=%x nif_no=%d nif_subno=%d\n", subcmd, nif_no, nif_subno);
				nr = NOE_NAK_INTERNAL;
				goto _error;
				
			}

			SET_ETHTOOL_OPS(netdev, &ops);

			memcpy(netdev->dev_addr, dev_changed->mac, sizeof(dev_changed->mac));

			INIT_WORK(&noe_priv->wqWirelessSendEvent, noept_wireless_send_event);
			INIT_WORK(&noe_priv->wqNetifUpdateFlags, noept_netif_update_flags);
			INIT_WORK(&noe_priv->wqUpdateNetdevFlags, noe_update_netdev_flags);
			INIT_WORK(&noe_priv->wqTxTimeout, noe_tx_timeout);

// don't attach to USB device tree to avoid registering udev database
//			SET_NETDEV_DEV(netdev, &noe_drv->interface->dev);
//

                        /* modified by sony */
                        if( rtnl_lock_already ){
                            /* not call dev_alloc_name. OK?? */
                            if(register_netdevice(netdev) != 0) {
                                NOE_ERROR("couldn't register the device!\n");
                                free_netdev(netdev);
                                nr = NOE_NAK_NO_RESOURCE;
                                goto _error;
                            }
                        }
                        else{
                            if(register_netdev(netdev) != 0) {
                                NOE_ERROR("couldn't register the device\n");
                                free_netdev(netdev);
                                nr = NOE_NAK_NO_RESOURCE;
                                goto _error;
                            }
                        }
			switch(subcmd) {
			case NOE_F_DCC_ADDED_ETH:
                                register_netdevinfo( noe_drv, nif_no, nif_subno, NOE_NETDEV_TYPE_ETHER, netdev );
				break;
			case NOE_F_DCC_ADDED_WIFI:
                                register_netdevinfo( noe_drv, nif_no, nif_subno, NOE_NETDEV_TYPE_WIFI, netdev );
				break;
			case NOE_F_DCC_ADDED_VAP:
                                register_netdevinfo( noe_drv, nif_no, nif_subno, NOE_NETDEV_TYPE_VAP, netdev );
				break;
                        default:
                                break;
                        }
			
			NOE_DP("%s: %s was created\n", __func__, netdev->name);
		}
	}
_error:;
	send_noe_response(noe_drv, nr, noe_drv->dc_ssid);
}

static void
rx_fixup(noe_drv_t *noe_drv)
{
	int status;
        static noe_rcv_usb_param param; /* added by sony */

        NOE_TRACE_BULK("rx_fixup\n");

        /************* modified by sony *******************/
        if( unlikely(in_atomic()) ){
            NOE_INFO("receive sequence is borken in atomic!!\n");
            panic(0);
            return;
        }

        if( test_and_set_bit(DEF_NFE_STATUS_NET_RECEIVING, &noe_drv->noe_status) ){
            NOE_INFO("receive sequence is borken!!\n");
            panic(0);
            return;
        }

        if( !is_init_usb_status() ){
            NOE_DP("clear INIT status. So not call usb_submit_urb \n");
            return;
        }

        if( _is_hdrl_rec_pause_condition( noe_drv ) ){
          /* schedule at check_rec_pause_timer sequence by sony */
          return;
        }

        param.noe_drv = noe_drv;
        param.data = _allocate_rx_buf(noe_drv);
        if( param.data == NULL ){
            NOE_INFO("cannot allocate buffer for Read URB\n");
            return;
        }
        /***************************************************/

        /* changed parameter by sony */
	usb_fill_bulk_urb(noe_drv->read_urb, noe_drv->udev,
                          usb_rcvbulkpipe(noe_drv->udev, noe_drv->bulk_in_endpointAddr/*1*/),
                          GET_RCV_USB_DATA_BODY(param.data), noe_drv->noe_packet_size_status.recv_payload_size,
                          ether_read_callback, &param); /* modified by sony */

        NOE_TRACE_BULK("wait receive from read_urb............\n");
	status = usb_submit_urb(noe_drv->read_urb, GFP_ATOMIC);
	if(status == -ENODEV) {
		int	i, j;
		for(i = 0;i < MAX_NIF;i++) {
			for(j = 0;j < MAX_NIF_SUB;j++) {
                          struct net_device *netdev = netdevinfo2netdev(noe_drv, i, j);
				if(netdev)
					netif_device_detach(netdev);
			}
		}
                _free_rx_buf(noe_drv, param.data); /* modified by sony */
	}
	else if(status){
                clear_bit(DEF_NFE_STATUS_NET_RECEIVING, &noe_drv->noe_status);
                _free_rx_buf(noe_drv, param.data); /* modified by sony */
                _request_rx_fixup(noe_drv);/* modified by sony */
        }
}

#define	WREQ_ALLOC_UNIT	512

void
ether_read_callback(struct urb *urb)
{
	noe_drv_t		*noe_drv;
	unsigned		pkt_len;
	NOE_LINK_STATUS		*noe_frame;
	NOE_UINT32		noe_cmd;
	NOE_NIFID	nif_id;
	struct net_device	*netdev;
	net_device_private_t	*noe_priv;
	int	l;
	NOE_UINT32		*p32;
	NOE_UINT32		t32;
        noe_rcv_usb_param       *rcv_param; /* added by sony */
        noe_rcv_usb_data        *rcv_data; /* added by sony */
        void                    *rcv_buffer; /* added by sony */

        NOE_TRACE_BULK("receive from read_urb............\n");
        /************* modified by sony *******************/
        rcv_param = urb->context; 
	noe_drv = rcv_param->noe_drv;
        rcv_data = rcv_param->data;
        rcv_buffer = GET_RCV_USB_DATA_BODY(rcv_data);
        clear_bit(DEF_NFE_STATUS_NET_RECEIVING, &noe_drv->noe_status);
        /***************************************************/

	if(noe_drv == NULL){
                NOE_ERROR("noe_drv is not set!!!\n");
		return;
        }
	if(test_bit(ETHER_UNPLUG, &noe_drv->flags)){
                _free_rx_buf(noe_drv, rcv_data); /* added by sony */
		return;
        }

        _validate_noe_system( noe_drv, urb->status, rcv_buffer ); /* added by sony */

	switch (urb->status) {
	case 0:
		break;
	case -ENOENT:
                _free_rx_buf(noe_drv, rcv_data); /* added by sony */
		return;	/* the urb is in unlink state */
	case -ETIME:
		NOE_INFO("may be reset is needed?..\n");
                _free_rx_buf(noe_drv, rcv_data); /* added by sony */
		goto goon;
	default:
		NOE_DP("Rx status %d\n", urb->status);
                _free_rx_buf(noe_drv, rcv_data); /* added by sony */
		goto goon;
	}

	/* protect against short packets */
	if(urb->actual_length < sizeof(NOE_HEADER)){
                _free_rx_buf(noe_drv, rcv_data); /* added by sony */
		goto goon;
        }

	pkt_len = urb->actual_length;

	//noe route
	noe_frame = (NOE_LINK_STATUS *)rcv_buffer; /* modified by sony */
	noe_cmd = sub2host32(noe_frame->h.cmd);
//NOE_DP("noe_cmd=%x", noe_cmd);

        ENTRY_CMD_HISTORY( noe_drv, DEBUG_HISTORY_TYPE_RECV, noe_cmd, noe_frame->h.len );

	if(noe_cmd == NOE_NETWORK_STATUS) {
          /*                NOE_DP("  NOE_NETWORK_STATUS packet received\n");*/
		if(sub2host32(noe_frame->h.len) > 0)
			noe_set_carrier(noe_drv, sub2host32(noe_frame->link_status));
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETWORK_STATUS);
                /*                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );*/
                _free_rx_buf(noe_drv, rcv_data);
	}
	else if(noe_cmd == NOE_NETWORK_PACKET) {
          /*                NOE_DP("  NOE_NETWORK_PACKET packet received\n");*/
                if( test_bit( DEF_NFE_STATUS_OPENED, &noe_drv->noe_status ) ){
                    noept_set_ether_blocks(noe_drv, rcv_buffer);
                }
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETWORK_PACKET);
                _free_rx_buf(noe_drv, rcv_data);
        }
	else if(noe_cmd == NOE_NETWORK_DEVICE_CHANGED) {
                NOE_DP("  NOE_NETWORK_DEVICE_CHANGED packet received\n");
                memcpy(&noe_drv->noe_dc, ((NOE_HEADER *)rcv_buffer) + 1, sizeof(NOE_DEVICE_CHANGED)); /* modified by sony */
		noe_drv->dc_ssid = sub2host32(noe_frame->h.SubSysID);
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_DEVICE_CHANGED);
                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );
                _free_rx_buf(noe_drv, rcv_data);
		queue_work(noe_drv->noe_event_dc_wq, &noe_drv->wqDevChanged);
	}
	else if(noe_cmd == NOE_BDEX_NETIF_SET_CARRIER) {
                NOE_DP("  NOE_BDEX_NETIF_SET_CARRIER packet received\n");
		nif_id = sub2host32(noe_frame->h.r0);
		netdev = netdevinfo2netdev( noe_drv, NOE_NIFID2NO(nif_id), NOE_NIFID2SUBNO(nif_id) );
		if(netdev) {
			if(noe_frame->h.r1/*no need to care about endian*/)
				netif_carrier_on(netdev);
			else
				netif_carrier_off(netdev);
		}
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETIF_SET_CARRIER);
                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );
                _free_rx_buf(noe_drv, rcv_data);
	}
	else if(noe_cmd == NOE_BDEX_NETIF_WIRELESS_SEND_EVENT_SIOCGIWAP) {
                NOE_DP("  NOE_BDEX_NETIF_WIRELESS_SEND_EVENT_SIOCGIWAP packet received\n");
		nif_id = sub2host32(noe_frame->h.r0);
		netdev = netdevinfo2netdev( noe_drv, NOE_NIFID2NO(nif_id), NOE_NIFID2SUBNO(nif_id) );
		if(netdev) {
		  //net_device_private_t *noe_priv;
			noe_priv = noe_netdev_priv(netdev);
			noe_priv->wreq_ssid = sub2host32(noe_frame->h.SubSysID);
			noe_priv->wreq_cmd = SIOCGIWAP;
			memset(&noe_priv->wreq, 0, sizeof(noe_priv->wreq));
			noe_priv->wreq.addr.sa_family = (sa_family_t)sub2host32(noe_frame->h.r1);
			memcpy(noe_priv->wreq.addr.sa_data, ((NOE_HEADER *)rcv_buffer) + 1, 6);/*modified by sony*/
			queue_work(noe_drv->noe_event_wq, &noe_priv->wqWirelessSendEvent);
		}
		else {
			NOE_DP("NOE_BDEX_NETIF_WIRELESS_SEND_EVENT_SIOCGIWAP requests for wrong netdev id(%x)\n", nif_id);
		}
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETIF_WIRELESS_SEND_EVENT_SIOCGIWAP);
                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );
                _free_rx_buf(noe_drv, rcv_data);
	}
	else if(noe_cmd == NOE_BDEX_NETIF_WIRELESS_SEND_EVENT) {
                NOE_DP("  NOE_BDEX_NETIF_WIRELESS_SEND_EVENT packet received\n");
		nif_id = sub2host32(noe_frame->h.r0);
		netdev = netdevinfo2netdev( noe_drv, NOE_NIFID2NO(nif_id), NOE_NIFID2SUBNO(nif_id) );
		if(netdev) {
		  //net_device_private_t *noe_priv;
			noe_priv = noe_netdev_priv(netdev);
			noe_priv->wreq_ssid = sub2host32(noe_frame->h.SubSysID);
			noe_priv->wreq_cmd = sub2host32(noe_frame->h.r1);
			memset(&noe_priv->wreq, 0, sizeof(noe_priv->wreq));
			noe_priv->wreq.data.length = sub2host32(noe_frame->h.len);
			noe_priv->wreq.data.flags = sub2host32(noe_frame->h.r2);
			if(noe_priv->wreq.data.length > noe_priv->wreq_buf_len) {
				if(noe_priv->wreq_buf)
					kfree(noe_priv->wreq_buf);
				noe_priv->wreq_buf_len = (noe_priv->wreq.data.length + WREQ_ALLOC_UNIT - 1) & ~(WREQ_ALLOC_UNIT-1);
				noe_priv->wreq_buf = kmalloc(noe_priv->wreq_buf_len, GFP_ATOMIC);
				if(noe_priv->wreq_buf == NULL) {
					NOE_ERROR("%s(%d): Out of memory\n", __func__, __LINE__);
					goto goon;
				}
			}
			if(noe_priv->wreq.data.length && noe_priv->wreq_buf)
                          memcpy(noe_priv->wreq_buf, ((NOE_HEADER *)rcv_buffer) + 1, noe_priv->wreq.data.length);/*modified by sony*/
			queue_work(noe_drv->noe_event_wq, &noe_priv->wqWirelessSendEvent);
		}
		else {
			NOE_DP("NOE_BDEX_NETIF_WIRELESS_SEND_EVENT requests for wrong netdev id(%x)\n", nif_id);
		}
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETIF_WIRELESS_SEND_EVENT);
                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );
                _free_rx_buf(noe_drv, rcv_data);
	}
	else if(noe_cmd == NOE_BDEX_NETIF_WIRELESS_SEND_EVENT_UINT32) {
                NOE_DP("  NOE_BDEX_NETIF_WIRELESS_SEND_EVENT_UINT32 packet received\n");
		nif_id = sub2host32(noe_frame->h.r0);
		netdev = netdevinfo2netdev( noe_drv, NOE_NIFID2NO(nif_id), NOE_NIFID2SUBNO(nif_id) );
		if(netdev) {
		  //net_device_private_t *noe_priv;
			noe_priv = noe_netdev_priv(netdev);
			noe_priv->wreq_ssid = sub2host32(noe_frame->h.SubSysID);
			noe_priv->wreq_cmd = sub2host32(noe_frame->h.r1);
			memset(&noe_priv->wreq, 0, sizeof(noe_priv->wreq));
			noe_priv->wreq.data.length = sub2host32(noe_frame->h.len);
			noe_priv->wreq.data.flags = sub2host32(noe_frame->h.r2);

			if(noe_priv->wreq.data.length > noe_priv->wreq_buf_len) {
				if(noe_priv->wreq_buf)
					kfree(noe_priv->wreq_buf);
				noe_priv->wreq_buf_len = (noe_priv->wreq.data.length + WREQ_ALLOC_UNIT - 1) & ~(WREQ_ALLOC_UNIT-1);
				noe_priv->wreq_buf = kmalloc(noe_priv->wreq_buf_len, GFP_ATOMIC);
				if(noe_priv->wreq_buf == NULL) {
					NOE_ERROR("%s(%d): Out of memory\n", __func__, __LINE__);
					goto goon;
				}
			}
			if(noe_priv->wreq.data.length && noe_priv->wreq_buf) {
				u_int32_t *p = (u_int32_t *)noe_priv->wreq_buf;
				u_int32_t *q = (u_int32_t *)(((NOE_HEADER *)rcv_buffer) + 1);/*modified by sony*/
				int	i;
				for(i = 0;i < noe_priv->wreq.data.length;i += 4) {
					*p++ = sub2host32(*q);
					q++;
				}
			}
			queue_work(noe_drv->noe_event_wq, &noe_priv->wqWirelessSendEvent);
		}
		else {
			NOE_DP("NOE_BDEX_NETIF_WIRELESS_SEND_EVENT requests for wrong netdev id(%x)\n", nif_id);
		}
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETIF_WIRELESS_SEND_EVENT_UINT32);
                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );
                _free_rx_buf(noe_drv, rcv_data);
	}
	else if(noe_cmd == NOE_BDEX_NETIF_UPDATE_FLAGS_S2H) {
                NOE_DP("  NOE_BDEX_NETIF_UPDATE_FLAGS_S2H packet received\n");
		nif_id = sub2host32(noe_frame->h.r0);
		netdev = netdevinfo2netdev( noe_drv, NOE_NIFID2NO(nif_id), NOE_NIFID2SUBNO(nif_id) );
		if(netdev) {
		  //net_device_private_t *noe_priv;
			noe_priv = noe_netdev_priv(netdev);
			noe_priv->netif_update_flags_ssid = sub2host32(noe_frame->h.SubSysID);
			noe_priv->netif_update_flags_and = sub2host32(noe_frame->h.r1);
			noe_priv->netif_update_flags_or = sub2host32(noe_frame->h.r2);
			queue_work(noe_drv->noe_event_wq, &noe_priv->wqNetifUpdateFlags);
		}
		else {
			NOE_DP("NOE_BDEX_NETIF_UPDATE_FLAGS_S2H requests for wrong netdev id(%x)\n", nif_id);
		}
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETIF_UPDATE_FLAGS_S2H);
                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );
                _free_rx_buf(noe_drv, rcv_data);
	}

#if defined(CONFIG_PREEMPT)
#define	NUM_ITEM_IN_STATS	24
#define	NUM_ITEM_IN_IW_STATS	9
	else if(noe_cmd == NOE_BDEX_NETIF_UPDATE_STATS) {
		l = sub2host32(noe_frame->h.len) / (sizeof(NOE_UINT32)*NUM_ITEM_IN_STATS);
		p32 = (NOE_UINT32 *)(((NOE_HEADER *)rcv_buffer) + 1); /*modeified by sony*/
	        NOE_DP("  NOE_BDEX_NETIF_UPDATE_STATS[%x] packet received\n", sub2host32(p32[0]));
		while(l--) {
			nif_id = sub2host32(p32[0]);
			netdev = netdevinfo2netdev( noe_drv, NOE_NIFID2NO(nif_id), NOE_NIFID2SUBNO(nif_id) );
                        if(netdev) {
				noe_priv = noe_netdev_priv(netdev);
				noe_priv->stats.rx_packets = sub2host32(p32[1]);
				noe_priv->stats.tx_packets = sub2host32(p32[2]);
				noe_priv->stats.rx_bytes = sub2host32(p32[3]);
				noe_priv->stats.tx_bytes = sub2host32(p32[4]);
				noe_priv->stats.rx_errors = sub2host32(p32[5]);
				noe_priv->stats.tx_errors = sub2host32(p32[6]);
				noe_priv->stats.rx_dropped = sub2host32(p32[7]);
				noe_priv->stats.tx_dropped = sub2host32(p32[8]);
				noe_priv->stats.multicast = sub2host32(p32[9]);
				noe_priv->stats.collisions = sub2host32(p32[10]);
				noe_priv->stats.rx_length_errors = sub2host32(p32[11]);
				noe_priv->stats.rx_over_errors = sub2host32(p32[12]);
				noe_priv->stats.rx_crc_errors = sub2host32(p32[13]);
				noe_priv->stats.rx_frame_errors = sub2host32(p32[14]);
				noe_priv->stats.rx_fifo_errors = sub2host32(p32[15]);
				noe_priv->stats.rx_missed_errors = sub2host32(p32[16]);
				noe_priv->stats.tx_aborted_errors = sub2host32(p32[17]);
				noe_priv->stats.tx_carrier_errors = sub2host32(p32[18]);
				noe_priv->stats.tx_fifo_errors = sub2host32(p32[19]);
				noe_priv->stats.tx_heartbeat_errors = sub2host32(p32[20]);
				noe_priv->stats.tx_window_errors = sub2host32(p32[21]);
				noe_priv->stats.rx_compressed = sub2host32(p32[22]);
				noe_priv->stats.tx_compressed = sub2host32(p32[23]);
				p32 += NUM_ITEM_IN_STATS;
			}
			else {
#if 0
                                int i,j;
				NOE_DP("  NOE_BDEX_NETIF_UPDATE_STATS requests for wrong netdev id(%x)\n", nif_id);
                                for( i = 0; i < MAX_NIF; i++ ){
                                  for( j = 0; j < MAX_NIF_SUB; j++ ){
                                    if( noe_drv->netdev[i][j] ){
                                      NOE_DP("  netdev[%d][%d] : %s\n", i, j, noe_drv->netdev[i][j]->name);
                                    }
                                  }
                                }
#endif
				break;
			}
		}
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETIF_UPDATE_STATS);
                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );
                _free_rx_buf(noe_drv, rcv_data);
	}

	else if(noe_cmd == NOE_BDEX_NETIF_UPDATE_IW_STATS) {
		l = sub2host32(noe_frame->h.len) / (sizeof(NOE_UINT32)*NUM_ITEM_IN_IW_STATS);
		p32 = (NOE_UINT32 *)(((NOE_HEADER *)rcv_buffer) + 1);/*modeified by sony*/
		while(l--) {
			nif_id = sub2host32(p32[0]);
			netdev = netdevinfo2netdev( noe_drv, NOE_NIFID2NO(nif_id), NOE_NIFID2SUBNO(nif_id) );
			if(netdev) {
				noe_priv = noe_netdev_priv(netdev);
				noe_priv->iw_stats.status = (__u16)sub2host32(p32[1]);
				t32 = sub2host32(p32[2]);
				noe_priv->iw_stats.qual.qual = (__u8)t32;
				noe_priv->iw_stats.qual.level = (__u8)(t32 >> 8);
				noe_priv->iw_stats.qual.noise = (__u8)(t32 >> 16);
				noe_priv->iw_stats.qual.updated = (__u8)(t32 >> 24);
				noe_priv->iw_stats.discard.nwid = sub2host32(p32[3]);
				noe_priv->iw_stats.discard.code = sub2host32(p32[4]);
				noe_priv->iw_stats.discard.fragment = sub2host32(p32[5]);
				noe_priv->iw_stats.discard.retries = sub2host32(p32[6]);
				noe_priv->iw_stats.discard.misc = sub2host32(p32[7]);
				noe_priv->iw_stats.miss.beacon = sub2host32(p32[8]);
				p32 += NUM_ITEM_IN_IW_STATS;
			}
			else {
				break;
			}
		}
                /*added by sony*/
                SET_RCV_USB_DATA_TYPE(rcv_data, NFE_RX_DATA_TYPE_NETIF_UPDATE_IW_STATS);
                DUMP_USB_PACKET( pkt_len, (NOE_HEADER*)noe_frame );
                _free_rx_buf(noe_drv, rcv_data);

	}

#endif
	else{
            noe_rec_data(urb, noe_drv, rcv_data); /* free buffer later in usb_drv.c */
        }

goon:;
        _request_rx_fixup(noe_drv);/* modified by sony */
}

/*
**
**	network related part of the code
**
*/


static struct net_device_stats *
ether_get_stats(struct net_device *netdev)
{
	//NOE_DP("stats:%x\n",(int)((noe_drv_t *)net2noe_drv(dev))->stats.tx_carrier_errors);
	return(&(noe_netdev_priv(netdev)->stats));
}

static void
ether_tx_timeout(struct net_device *netdev)
{
	NOE_INFO("%s: Tx timeout.\n", netdev->name);

	noe_netdev_priv(netdev)->stats.tx_errors++;
}

static void
ether_write_callback(struct urb *urb)
{
	noe_drv_t	 	*noe_drv;
	struct net_device	*netdev;

	NOE_DP("write_bulk_callback\n");

	netdev = urb->context;
	if(netdev == NULL){
            panic("no netdev at %s\n", __FUNCTION__);
            return;
        }

	noe_drv = net2noe_drv(netdev);
	dev_kfree_skb_irq(noe_drv->tx_skb);
	if(!netif_device_present(netdev)) {
            panic("no noe_drv at %s\n", __FUNCTION__);
            return;
	}

	if(urb->status)
		NOE_DP("%s: Tx status %d\n", netdev->name, urb->status);
	netdev->trans_start = jiffies + NOE_TIMEOUT;
	netif_wake_queue(netdev);

/* add by sony */
        clear_bit(RIGHT_TO_USE_BULKOUT, &noe_drv->flags);
        if(waitqueue_active( &noe_drv->bulkout_wait )){
            wake_up_interruptible( &noe_drv->bulkout_wait );
        }

        return;
}

int
ether_hard_start_xmit(struct sk_buff *skb, struct net_device *netdev)
{
	net_device_private_t	*noe_priv;
	noe_drv_t		*noe_drv;
	int			retval;
	NOE_HEADER		*noe_header;
	struct sk_buff		*skb2;
	unsigned int		packet_len;

	noe_priv = noe_netdev_priv(netdev);
	noe_drv = noe_priv->noe_drv;

        //	NOE_DP("ether_start_xmit\n");

#ifdef NOE_SACK_CHECK_MODE
        check_sack(skb);
#endif

	if(test_and_set_bit(RIGHT_TO_USE_BULKOUT, &noe_drv->flags)) {
		// someone has right to use bulk out pipe.
		// ask reschedule skb
		return(NETDEV_TX_BUSY);
	}

	netif_stop_queue(netdev);

	packet_len = skb->len;

        //        DUMP_SKB(skb);

	/* Create a new skb to use with the correct size */
	skb2 = skb_copy_expand(skb, sizeof(NOE_HEADER), 0, GFP_ATOMIC);
	dev_kfree_skb_any(skb);
	if(skb2 == NULL)
		goto _path_send_data_done;

	//add noe_header
	noe_header = (NOE_HEADER *)skb_push(skb2, sizeof(NOE_HEADER));
	noe_header->HostID = host2sub32(NOE_HID_UNBIND);
	noe_header->SubSysID = host2sub32(NOE_SID_UNBIND);
	noe_header->len = host2sub32(packet_len);
	noe_header->cmd = host2sub32(NOE_NETWORK_PACKET);
	noe_header->ctrl = host2sub32(noe_priv->nif_id);

	noe_drv->tx_skb = skb2;
	packet_len = skb2->len;			// skb2 might be freed after returns from usb_submit_urb

        if( test_bit(DEF_NFE_DEBUG_NETWORK_MONITOR, &noe_drv->debug_status) ){
          noe_drv->debug_pass_thru_snd_len += packet_len;
        }

	usb_fill_bulk_urb(noe_drv->write_urb, noe_drv->udev, usb_sndbulkpipe(noe_drv->udev, noe_drv->bulk_out_endpointAddr/*2*/),
						skb2->data, NOE_MAKEALIGN(packet_len), ether_write_callback, netdev);
	//support zero length packet
	noe_drv->write_urb->transfer_flags |= URB_ZERO_PACKET;
	if((retval = usb_submit_urb(noe_drv->write_urb, GFP_ATOMIC))) {
		/* Can we get/handle EPIPE here? */
		if(retval == -ENODEV)
			netif_device_detach(netdev);
		else {
			NOE_ERROR("%s: failed write_urb %d\n", __func__, retval);
			noe_priv->stats.tx_errors++;
			netif_start_queue(netdev);
		}

                goto _path_send_data_done;
	}
	else {
		noe_priv->stats.tx_packets++;
		noe_priv->stats.tx_bytes += packet_len;
		netdev->trans_start = jiffies + NOE_TIMEOUT;
	}
	return(0);

/* add by sony */
_path_send_data_done:;

        clear_bit(RIGHT_TO_USE_BULKOUT, &noe_drv->flags);
        if(waitqueue_active( &noe_drv->bulkout_wait )){
            wake_up_interruptible( &noe_drv->bulkout_wait );
        }

        return(0);
}

static int
ether_open(struct net_device *netdev)
{
	noe_drv_t *noe_drv = net2noe_drv(netdev);

	NOE_DP("ether_open:%s\n", netdev->name);

	atomic_inc(&(noe_netdev_priv(netdev)->opened));
	atomic_inc_return(&noe_drv->ether_opened);

        set_bit( DEF_NFE_STATUS_OPENED, &noe_drv->noe_status ); /* added by sony */

	set_carrier(netdev);
	netif_start_queue(netdev);

	return(0);
}

static int
ether_close(struct net_device *netdev)
{
	noe_drv_t *noe_drv = net2noe_drv(netdev);

        clear_bit( DEF_NFE_STATUS_OPENED, &noe_drv->noe_status ); /* added by sony */

	atomic_dec(&noe_drv->ether_opened);
	atomic_dec(&(noe_netdev_priv(netdev)->opened));
	netif_stop_queue(netdev);

	return(0);
}

static void
ether_get_drvinfo(struct net_device *netdev, struct ethtool_drvinfo *info)
{
	noe_drv_t *noe_drv = net2noe_drv(netdev);

	strncpy(info->driver, driver_name, ETHTOOL_BUSINFO_LEN);
	strncpy(info->version, driver_version, ETHTOOL_BUSINFO_LEN);
	usb_make_path(noe_drv->udev, info->bus_info, sizeof info->bus_info);
}

static int
ether_get_settings(struct net_device *netdev, struct ethtool_cmd *ecmd)
{
	ecmd->supported = (SUPPORTED_10baseT_Half |
				  SUPPORTED_10baseT_Full |
				  SUPPORTED_100baseT_Half |
				  SUPPORTED_100baseT_Full |
				  SUPPORTED_Autoneg | SUPPORTED_TP);
	ecmd->port = PORT_TP;
	ecmd->transceiver = XCVR_INTERNAL;
	ecmd->phy_address = 0;

	ecmd->autoneg = AUTONEG_ENABLE;
	ecmd->speed = SPEED_100;
	ecmd->duplex = DUPLEX_FULL;

	return(0);
}

int
ether_probe(noe_drv_t *noe_drv)
{
	int	retval = -ENOMEM;

        NOE_DP("NOE ethernet probed.......\n");

        /* modified by sony */
        /*  The processing thread change to work queue from tasklet */
        atomic_set(&noe_drv->rx_fixup_req_cnt, 0);
        init_waitqueue_head(&noe_drv->rx_fixup_wait_queue);

        noe_drv->noe_rx_fixup_thread = _create_rx_fixup_thread(noe_drv);
        if( noe_drv->noe_rx_fixup_thread == NULL ){
		goto _error;
        }

        /* modified by sony */
        /*   The buffer changes to by kmalloc from by kmem_cache_create */
        noe_drv->rcv_buffer = kmem_cache_create("noe_rcv_cache",
                                                (noe_drv->noe_packet_size_status.recv_payload_size)+sizeof(noe_rcv_usb_attr),
                                                0, SLAB_HWCACHE_ALIGN, NULL);
	if(noe_drv->rcv_buffer == NULL) {
		NOE_ERROR("couldn't alloc rcv_buffer\n");
		goto _error;
	}
#ifdef CONFIG_NOE_DEBUG_MEMORY
        INIT_LIST_HEAD(&noe_drv->rcv_buffer_dbg_list);
#endif

	noe_drv->noe_event_dc_wq = create_singlethread_workqueue("noe_event_dc");
	if(noe_drv->noe_event_dc_wq == NULL) {
		NOE_ERROR("couldn't create work queue DC\n");
		goto _error;
	}
	noe_drv->noe_event_wq = create_singlethread_workqueue("noe_event");
	if(noe_drv->noe_event_wq == NULL) {
		NOE_ERROR("couldn't create work queue\n");
		goto _error;
	}

	INIT_WORK(&noe_drv->wqDevChanged, noept_device_changed);
	noe_drv->num_nif = get_num_nif(noe_drv);
        _request_rx_fixup(noe_drv); /* modified by sony */

#if defined(CONFIG_PREEMPT)
					// don't call more than once
	set_registers(noe_drv, set_Net_Stats_Period, 1000, 0, NULL);
#endif

        /********* added by sony *********/
	init_timer(&noe_drv->check_rec_pause_timer);
	noe_drv->check_rec_pause_timer.function = _check_rec_pause;
	noe_drv->check_rec_pause_timer.data = (unsigned long)noe_drv;


	return(0);

_error:;
	if(noe_drv->noe_event_wq) {
		destroy_workqueue(noe_drv->noe_event_wq);
		noe_drv->noe_event_wq = NULL;
	}
	if(noe_drv->noe_event_dc_wq) {
		destroy_workqueue(noe_drv->noe_event_dc_wq);
		noe_drv->noe_event_dc_wq = NULL;
	}
	if(noe_drv->rcv_buffer) {/* modified by sony */
                kmem_cache_destroy(noe_drv->rcv_buffer);
		noe_drv->rcv_buffer = NULL;
	}
        if(noe_drv->noe_rx_fixup_thread) {/* modified by sony */
                _destroy_rx_fixup_thread( noe_drv );
        }

	return(retval);
}

void
ether_disconnect(struct usb_interface *interface)
{
	int	i, j;
	noe_drv_t *noe_drv = usb_get_intfdata(interface);

	NOE_DP("ether_disconnect\n");

	if(noe_drv) {
		set_bit(ETHER_UNPLUG, &noe_drv->flags);
                /* modified by sony */
                NOE_DP(" destory rx_fixup thread.........\n");
                _destroy_rx_fixup_thread( noe_drv );
                NOE_DP(" destory rx_fixup thread.........done\n");
                NOE_DP(" unregister net device.....\n");

		for(i = 0;i < MAX_NIF;i++) {
			for(j = 0;j < MAX_NIF_SUB;j++) {
                                struct net_device *netdev;
                                netdev = netdevinfo2netdev( noe_drv, i, j );
				if(netdev) {
					NOE_DP("%s: ether_disconnect\n", netdev->name);
					unregister_netdevinfo( noe_drv, i, j );
					unregister_netdev(netdev);
					free_netdev(netdev);
				}
			}
		}
		if(noe_drv->noe_event_wq) {
                        NOE_DP(" destory workqueue:noe_event.......\n");
			flush_workqueue(noe_drv->noe_event_wq);
			destroy_workqueue(noe_drv->noe_event_wq);
			noe_drv->noe_event_wq = NULL;
		}
		if(noe_drv->noe_event_dc_wq) {
                        NOE_DP(" destory workqueue:noe_event_dc.......\n");
			/* ToDo */
			flush_workqueue(noe_drv->noe_event_dc_wq);
			destroy_workqueue(noe_drv->noe_event_dc_wq);
			noe_drv->noe_event_dc_wq = NULL;
		}
		if(noe_drv->rcv_buffer) {
                        NOE_DP(" destory rcv_buffer.......\n");
                        kmem_cache_destroy(noe_drv->rcv_buffer); /* modified by sony */
			noe_drv->rcv_buffer = NULL;
#ifdef CONFIG_NOE_DEBUG_MEMORY
                        _check_rcv_buffer(noe_drv);
#endif

		}

                /* added by sony */
                _clear_check_hdrl_rec_pause_timer(noe_drv);
		
	}
}

/***************************************************/
/*           new functions by SONY                 */
/***************************************************/
/* Internal Functions */
static void _validate_noe_system( noe_drv_t *noe_drv, int status, void *buffer )
{
    static int status_error_cnt = 0;
#define DEF_NFE_NET_STATUS_ERR_CNT 30

    if( status == 0 ){
        status_error_cnt = 0;
    }
    else{
        status_error_cnt++;

        if( test_bit(DEF_NFE_STATUS_DO_SLEEP_FLAG, &noe_drv->noe_status) ){
            NOE_DP("URB Error due to set RECEIVE_DO_SLEEP_FLAG. So not call usb_submit_urb \n");
            return;
        }

        if( !is_init_usb_status() ){
            if( status == -ESHUTDOWN ){
                NOE_DP("URB Error due to clear INIT status for Shutdown.\n");
            }
            else{
                NOE_INFO("URB Error due to clear INIT status:%d.\n", status);
            }
            return;
        }

        if( status_error_cnt < 10 || ( status_error_cnt % 10 ) == 0 ){
            NOE_INFO("URB Error: Rx status %d [cnt:%d]<%08x>\n", status, status_error_cnt,(unsigned int)jiffies);
        }

        if( status_error_cnt > DEF_NFE_NET_STATUS_ERR_CNT ){
            PRINT_CMD_HISTORY(noe_drv,DEBUG_HISTORY_TYPE_ALL);
            NOE_ERROR("##########################\n");
            NOE_ERROR("##########################\n");
            NOE_ERROR("noe panic\n");
            NOE_ERROR("##########################\n");
            NOE_ERROR("##########################\n");
            panic("noe_panic");
        }
    }
}

/* Read Bulk Control */
static struct task_struct* _create_rx_fixup_thread( noe_drv_t *noe_drv )
{
    struct task_struct *thread;
    struct sched_param sched_param;

    init_completion( &noe_drv->rx_fixup_thread_start_complete );
    init_completion( &noe_drv->rx_fixup_thread_finish_complete );
    set_bit( DEF_NFE_THREAD_STATUS_RX_RUNNING, &noe_drv->noe_thread_status );

    thread = kthread_run( _rx_fixup_thread_main, noe_drv, "noe_rx_fixup" );
    if( thread == NULL ){
        NOE_ERROR("couldn't create rx_fixup_thread\n");
        return NULL;
    }

    wait_for_completion( &noe_drv->rx_fixup_thread_start_complete );

    sched_param.sched_priority = NOE_RX_FIXUP_THREAD_PRIORITY;
    if( sched_setscheduler( thread, NOE_RX_FIXUP_THREAD_POLICY, &sched_param ) < 0 ){
        NOE_ERROR("couldn't change priority for rx_fixup_thread\n");
        _destroy_rx_fixup_thread( noe_drv );
        return NULL;
    }

    return thread;
}

static void _destroy_rx_fixup_thread( noe_drv_t *noe_drv )
{
    clear_bit( DEF_NFE_THREAD_STATUS_RX_RUNNING, &noe_drv->noe_thread_status );
    _request_rx_fixup( noe_drv ); /* dummy request */
    wait_for_completion( &noe_drv->rx_fixup_thread_finish_complete );
}

static void _request_rx_fixup( noe_drv_t *noe_drv )
{
    atomic_inc(&noe_drv->rx_fixup_req_cnt);

    if( waitqueue_active(&noe_drv->rx_fixup_wait_queue) ){
        wake_up_interruptible(&noe_drv->rx_fixup_wait_queue);
    }
}

static int _rx_fixup_thread_main(void* data)
{
    noe_drv_t *noe_drv = (noe_drv_t*)data;

    complete( &noe_drv->rx_fixup_thread_start_complete );

    while(1){
      int cnt;

        wait_event_interruptible( noe_drv->rx_fixup_wait_queue, (atomic_read(&noe_drv->rx_fixup_req_cnt) > 0) );

        cnt = atomic_dec_return(&noe_drv->rx_fixup_req_cnt);

        if( !test_bit( DEF_NFE_THREAD_STATUS_RX_RUNNING, &(noe_drv->noe_thread_status) ) ){

            if( cnt > 0 ){
                NOE_INFO("cancel %d requests for Bulk Read operation\n", cnt);
            }
            break;
        }
        rx_fixup( noe_drv );
    }

    complete( &noe_drv->rx_fixup_thread_finish_complete );

    return 0;
}

static inline noe_rcv_usb_data* _allocate_rx_buf( noe_drv_t *noe_drv )
{
    noe_rcv_usb_data* data;

    data = (noe_rcv_usb_data*)kmem_cache_alloc( noe_drv->rcv_buffer, GFP_KERNEL );
    data->attr.type = NFE_RX_DATA_TYPE_UNKNOWN;
#ifdef CONFIG_NOE_DEBUG_MEMORY
    if( data != NULL ){
        list_add_tail( &data->attr.list, &noe_drv->rcv_buffer_dbg_list );
    }
#endif
    return data;
}

static inline void _free_rx_buf( noe_drv_t *noe_drv, noe_rcv_usb_data *data )
{
#ifdef CONFIG_NOE_DEBUG_MEMORY
    noe_rcv_usb_attr *c, *n;

    list_for_each_entry_safe( c, n, &noe_drv->rcv_buffer_dbg_list, list ){
        if( (noe_rcv_usb_data*)c == data ){
            list_del(&c->list);
            break;
        }
    }
#endif
    kmem_cache_free( noe_drv->rcv_buffer, data );
}

#ifdef CONFIG_NOE_DEBUG_MEMORY
static void _check_rcv_buffer( noe_drv_t *noe_drv )
{
    noe_rcv_usb_attr *c;
    int latest_cnt[NFE_RX_DATA_TYPE_MAX];
    int latest_total_cnt = 0;
    int i;

    for( i = 0; i < NFE_RX_DATA_TYPE_MAX; i++ ){
        latest_cnt[i] = 0;
    }

    list_for_each_entry( c, &noe_drv->rcv_buffer_dbg_list, list ){
        latest_cnt[c->type]++;
        latest_total_cnt++;
    }

    printk("#### Rx Data Buffer ####\n");
    for( i = 0; i < NFE_RX_DATA_TYPE_MAX; i++ ){
        printk("[%02d] %d\n", i, latest_cnt[i]);
    }
    printk("total:%d\n", latest_total_cnt);
}
#endif

/* REC Pause Control */
#define DEF_NOE_REC_PAUSE_JIFFIES ((12*HZ)/1000)  //HZ=250 12msec
static void _add_check_hdrl_rec_pause_timer( noe_drv_t *noe_drv )
{
    noe_drv->check_rec_pause_timer.expires = jiffies + DEF_NOE_REC_PAUSE_JIFFIES;
    add_timer(&noe_drv->check_rec_pause_timer);
}

static void _clear_check_hdrl_rec_pause_timer( noe_drv_t *noe_drv )
{
    del_timer_sync(&noe_drv->check_rec_pause_timer);
}

static void _check_rec_pause( unsigned long data )
{
    noe_drv_t *noe_drv = (noe_drv_t*)data;

    if( _is_hdrl_rec_pause_condition( noe_drv ) ){
        return;
    }
    _request_rx_fixup(noe_drv);
}

static bool _is_hdrl_rec_pause_condition( noe_drv_t *noe_drv )
{
    if( test_bit( DEF_NFE_REC_STATUS_REC_PAUSE, &noe_drv->noe_rec_status ) ){
        _add_check_hdrl_rec_pause_timer( noe_drv );
        return true;
    }
    return false;
}

/* call from usb_drv.c */
void net_drv_free_rx_buf( noe_drv_t *noe_drv, noe_rcv_usb_data *data )
{
    _free_rx_buf( noe_drv, data );
}

void net_drv_do_sleep_mode( noe_drv_t *noe_drv, bool enable )
{
    if( enable ){
        set_bit(DEF_NFE_STATUS_DO_SLEEP_FLAG, &noe_drv->noe_status);
    }
    else{
        clear_bit(DEF_NFE_STATUS_DO_SLEEP_FLAG, &noe_drv->noe_status);
    }
}

