//
//	$Revision: #1 $	$Date: 2012/08/03 $
//	Copyright (C) Ubiquitous Corp. 2001-2010. All rights reserved.
//

#ifndef	__NOE_DEV_H__
#define	__NOE_DEV_H__

/* by sony */
#include <linux/slab.h>
#include <linux/completion.h>
#include <linux/netdevice.h>
#include <linux/wireless.h>
#include <linux/usb.h>
#include <asm/semaphore.h>

#ifndef info
#define	info(format, arg...)	printk(KERN_INFO "%s: " format "\n", __FILE__, ##arg)
#endif
#ifndef warn
#define	warn(format, arg...)	printk(KERN_WARNING "%s: " format "\n", __FILE__, ##arg)
#endif
#ifndef err
#define	err(format, arg...)	printk(KERN_ERR "%s: " format "\n", __FILE__, ##arg)
#endif

#include "noe_std.h"
#include "noe_driver_if.h" /* by sony */

#define	MAX_NIF		4
#define	MAX_NIF_SUB	4


#define WQ_RINGS	16
#define NOE_TIMEOUT			(20*HZ)
#define SLEEP_WAIT_TIMEOUT			(5*HZ)
#define NOE_RINGS			64 //ring_buffers
#define	RX_SKB_POOL_SIZE	4

#define NOE_RING_NEAR_FULL	(NOE_RINGS-2)
#define NOE_RING_OK			0
#define NOE_RING_FULL		1

#define NOEPT_NUM_OF_ETHER_BLOCKS	(16)
#define NOEPT_ETHER_MTU				(1568)
#define NOE_ETHER_BLOCK_LENGTH		(1536)
#define NOE_ETHER_BLOCK_PAYLOAD		(NOE_ETHER_BLOCK_LENGTH - sizeof(NOE_UINT32))

typedef	struct {
	NOE_UINT32      	HostID;
	wait_queue_head_t	ioctl_wait;
} noe_wait_queue_t;

/*****************************************
  new define by sony
*****************************************/
/* kernel thread priority */
#define NOE_RX_FIXUP_THREAD_PRIORITY   90
#define NOE_RX_FIXUP_THREAD_POLICY     SCHED_RR

#define NOE_WRITE_OBJECT_THREAD_PRIORITY   90
#define NOE_WRITE_OBJECT_THREAD_POLICY     SCHED_RR

#undef CONFIG_NOE_DEBUG_MEMORY

/* rx data */
typedef enum {
  NFE_RX_DATA_TYPE_UNKNOWN = 0,
  NFE_RX_DATA_TYPE_NETWORK_STATUS,
  NFE_RX_DATA_TYPE_NETWORK_PACKET,
  NFE_RX_DATA_TYPE_DEVICE_CHANGED,
  NFE_RX_DATA_TYPE_NETIF_SET_CARRIER,
  NFE_RX_DATA_TYPE_NETIF_WIRELESS_SEND_EVENT_SIOCGIWAP,
  NFE_RX_DATA_TYPE_NETIF_WIRELESS_SEND_EVENT,
  NFE_RX_DATA_TYPE_NETIF_WIRELESS_SEND_EVENT_UINT32,
  NFE_RX_DATA_TYPE_NETIF_UPDATE_FLAGS_S2H,
  NFE_RX_DATA_TYPE_NETIF_UPDATE_STATS,
  NFE_RX_DATA_TYPE_NETIF_UPDATE_IW_STATS,
  NFE_RX_DATA_TYPE_WRITE_OBJECT,
  NFE_RX_DATA_TYPE_RCV_THD_DATA,
  NFE_RX_DATA_TYPE_OTHER,
  NFE_RX_DATA_TYPE_MAX,
} NFE_RX_DATA_TYPE;

typedef struct {
  NFE_RX_DATA_TYPE  type;
#ifdef CONFIG_NOE_DEBUG_MEMORY
  struct list_head  list;
#endif
} noe_rcv_usb_attr;

typedef struct {
  noe_rcv_usb_attr  attr; /* need top */
  /* contain body:max size is HOST_DRIVER_MAXPACKET */
} noe_rcv_usb_data;

#define GET_RCV_USB_DATA_BODY(data) ((NOE_pFRAME*)((void*)data + sizeof(noe_rcv_usb_attr)))
#define SET_RCV_USB_DATA_TYPE(data, type_name)  {data->attr.type = type_name;}

/* noe_status */
enum {
  DEF_NFE_STATUS_OPENED,
  DEF_NFE_STATUS_DO_SLEEP_FLAG,
  DEF_NFE_STATUS_PRE_CLOSE,
  DEF_NFE_STATUS_NET_RECEIVING,
};


/* noe_send_status */
enum {
  DEF_NFE_SEND_NOE_DATA,
};

/* noe_rec_status */
enum {
  DEF_NFE_REC_STATUS_REC_PAUSE,
};

/* noe_thread_status */
enum {
  DEF_NFE_THREAD_STATUS_RX_RUNNING,
};

/* debug_status */
enum {
  DEF_NFE_DEBUG_NETWORK_MONITOR,
  DEF_NFE_DEBUG_CHECK_STREAM,
};

/* do_sleep_flag */
enum {
  DEF_NFE_SLEEP_MODE,
  DEF_NFE_SLEEP_RECEIVE_PACKET,
};

#define NOE_INVALID_HOST_ID     -1 /* added by sony */

/*****************************************
  modified structure by sony
*****************************************/
typedef struct {
	NOE_UINT32	   HostID;
	int		   length;
        noe_rcv_usb_data  *data; /* modified by sony */
        struct list_head   list; /* modified by sony */
} noe_packet_buffer_t;

typedef struct {
        struct kmem_cache *cache;
        spinlock_t         list_lock;
        struct list_head   list;
        atomic_t           cnt;
        wait_queue_head_t  wq;
} hdrl_write_queue;

/* add for HDRL_Move */
typedef struct {
        struct kmem_cache *cache;
        spinlock_t         list_lock;
        struct list_head   ready_list;
        wait_queue_head_t  ready_wq;
        atomic_t           ready_cnt;
        struct list_head   fill_list;
        wait_queue_head_t  fill_wq;
        atomic_t           fill_cnt;
} request_write_queue;

typedef enum {
  NOE_HDRL_STATUS_IDLE,
  NOE_HDRL_STATUS_READY,
  NOE_HDRL_STATUS_TERMINATING,
} NOE_HDRL_STATUS;
/* */

typedef struct {
        struct list_head   used_list;
        struct list_head   free_list;
#ifdef DEBUG_PACKET_BUFFER_LIST
        int                used_cnt;
        int                free_cnt;
#endif
} noe_packet_buffer_list_t;

#ifdef DEBUG_NOE_COMMAND_TRACE
typedef struct {
        NOE_UINT32         cmd;
        NOE_UINT32         len;
} noe_packet_history_t;

typedef enum {
  DEBUG_HISTORY_TYPE_RECV = 0,
  DEBUG_HISTORY_TYPE_SEND,
  DEBUG_HISTORY_TYPE_MAX,
  DEBUG_HISTORY_TYPE_ALL,
} DEBUG_HISTORY_TYPE;

#define DEBUG_HISTORY_CNT_MAX       0x0000000f /*16*/
#endif

typedef enum {
    NOE_DEVICE_TYPE_UNKNOWN = -1,
    NOE_DEVICE_TYPE_ETHER = 0,
    NOE_DEVICE_TYPE_WIFI,
    NOE_DEVICE_TYPE_VAP,
    NOE_DEVICE_TYPE_MAX,
} noe_device_type;

enum {
    VAP_STATUS_DELETE_PROCESSING_BY_IOCTL,
};

struct netdev_info {
        struct list_head         list;
        struct net_device       *netdev;
        noe_device_type          type;
        NOE_UINT8                nif;
        NOE_UINT8                nif_sub;
};

/* Structure to hold all of our device specific stuff */
typedef struct {
	//usb
	struct usb_device       *udev;			/* the usb device for this device */
	struct usb_interface    *interface;		/* the interface for this device */

	struct urb		*read_urb;		/* the urb used to receive data */
	__u8			bulk_in_endpointAddr;	/* the address of the bulk in endpoint */

	unsigned char		*bulk_out_buffer;	/* the buffer to send data */
	struct urb		*write_urb;		/* the urb used to send data */
	__u8			bulk_out_endpointAddr;	/* the address of the bulk out endpoint */

	//schedule
	struct workqueue_struct	*noe_event_dc_wq;
	struct workqueue_struct	*noe_event_wq;
	struct kref		kref;
	struct work_struct	wqDevChanged;
	NOE_DEVICE_CHANGED	noe_dc;
	NOE_UINT32		dc_ssid;

	//network depend parts
	unsigned long		flags;
        struct netdev_info      netdev_info[MAX_NIF][MAX_NIF_SUB]; /* modified by sony */
	struct urb		*ctrl_urb;
        struct kmem_cache	*rcv_buffer; /* modified by sony kmalloc->kmem_cache */
#ifdef CONFIG_NOE_DEBUG_MEMORY
        struct list_head         rcv_buffer_dbg_list;
#endif
	
	//noe depend parts
	struct sk_buff		*tx_skb;

	// multi_nif
	unsigned int		num_nif;

	atomic_t		ether_opened;

	// bulk out
	wait_queue_head_t	bulkout_wait;

	// noe ioctl write
	struct mutex		write_mutex;
	wait_queue_head_t	ioctl_noe;

	// noe ioctl read
	spinlock_t		rd_ioctl_lock;
  //	noe_packet_buffer_t	packet_buffer[NOE_RINGS];
	noe_wait_queue_t	wait_queue[WQ_RINGS];

        /*************** added by sony *************/
        unsigned long           noe_status;
        unsigned long           noe_thread_status;

        struct task_struct     *noe_rx_fixup_thread; /* change from tasklet:tlEther */
        struct completion       rx_fixup_thread_start_complete;
        struct completion       rx_fixup_thread_finish_complete;
        wait_queue_head_t       rx_fixup_wait_queue;
        atomic_t                rx_fixup_req_cnt;

        unsigned long           noe_rec_status;
        struct timer_list       check_rec_pause_timer;

        noe_packet_buffer_list_t packet_buffer_by_hostid;

        unsigned long           debug_status;
        unsigned long           debug_pass_thru_rcv_len;
        unsigned long           debug_pass_thru_snd_len;
        unsigned long           debug_dlna_rcv_len;
        unsigned long           debug_noe_data_snd_len;
        int                     debug_hdrl_write_notify_cnt;
#ifdef DEBUG_NOE_COMMAND_TRACE
        noe_packet_history_t    debug_cmd_history[DEBUG_HISTORY_TYPE_MAX][DEBUG_HISTORY_CNT_MAX];
        atomic_t                debug_cmd_history_p[DEBUG_HISTORY_TYPE_MAX];
#endif

        struct semaphore        noe_rcv_thd_data_sem;
        struct list_head        noe_rcv_thd_data_list;
        spinlock_t              noe_rcv_thd_data_list_lock;
        atomic_t                noe_rcv_thd_data_list_cnt;
        wait_queue_head_t       noe_rcv_thd_data_wait_queue;
        struct kmem_cache      *noe_rcv_thd_buffer;

        wait_queue_head_t       do_sleep_wait_queue;
        unsigned long           do_sleep_flag;
        NOE_UINT32              do_sleep_hostid;

	/********************************/
        /*     HDRL                     */
	/********************************/
        struct task_struct     *write_object_thread;
        struct completion       write_object_thread_start_complete;
        struct completion       write_object_thread_finish_complete;

        hdrl_write_queue        hdrl_write_queue;
        atomic_t                hdrl_write_host_id; /* HostID = NOE_UINT32 */

        request_write_queue     request_write_queue;
        atomic_t                hdrl_status;




	unsigned char	       *read_obj_out_buffer;
        
        unsigned long           noe_send_status; /* <- send_noe_on */
        struct noe_dlna_highmem_info_st  dlna_highmem_info;
        struct noe_packet_size_status_st noe_packet_size_status;

        // for VAP status
        unsigned long           vap_status;

        // for NOE NIF 
        atomic_t                current_nif;
        // for NOE console
        NOE_UINT32              current_console_para;
        // for Debug
        NOE_UINT32              latest_noe_cmd;

        struct semaphore        ioctl_sem;

        /*******************************************/

} noe_drv_t;

/* noe_drv_t bitwised flags */
#define	ETHER_UNPLUG		2
#define	RIGHT_TO_USE_BULKOUT	4
#define	DRIVER_OPENED		5

typedef struct {
	struct net_device_stats	stats;
	struct iw_statistics	iw_stats;
	noe_drv_t		*noe_drv;
	atomic_t		opened;
	NOE_NIFID		nif_id;
	NOE_UINT32		dev_ssid;
	struct net_device	*netdev;

	// for SIOCGIWAP
	struct work_struct	wqWirelessSendEvent;
	NOE_UINT32		wreq_ssid;
	NOE_UINT32		wreq_cmd;
	union iwreq_data	wreq;
	int			wreq_buf_len;
	char			*wreq_buf;

	// for NETIF update flags
	struct work_struct	wqNetifUpdateFlags;
	NOE_UINT32		netif_update_flags_ssid;
	NOE_UINT32		netif_update_flags_and;
	NOE_UINT32		netif_update_flags_or;

	// to psss net_device->flags to subsystem
	struct work_struct	wqUpdateNetdevFlags;
	// to call subsystem's tx_timeout
	struct work_struct	wqTxTimeout;
} net_device_private_t;

extern int noe_probe(struct usb_interface *, const struct usb_device_id *);
extern void noe_disconnect(struct usb_interface *);
void ether_read_callback(struct urb *urb);
int ether_probe(noe_drv_t *dev);
void ether_disconnect(struct usb_interface *interface);

void	NOEki_send_noe_header_payload(noe_drv_t *dev, NOE_HEADER *h, void *payload);
int	NOEki_wait_normal_packet(noe_drv_t *dev, NOE_HEADER *h, void *payload);

#define	noe_netdev_priv(net_dev)	((net_device_private_t *)netdev_priv(net_dev))
#define	net2noe_drv(net_dev)		(noe_netdev_priv(net_dev)->noe_drv)

#define get_current_netdev(dev)         dev->netdev_info[(NOE_UINT32)atomic_read(&dev->current_nif)][0].netdev

#ifndef	SET_MODULE_OWNER
#define	SET_MODULE_OWNER(x)
#endif

#if defined(NOE_DEBUG)
#define	TOSTR(x)	#x
#define	MYASSERT(x)	do { if(!(x)) err("assert ", __func__, TOSTR(__LINE__)); } while(0)
#else
#define	MYASSERT(x)
#endif

#define	ETHER_MTU		1500
//#define	ETHER_TX_TIMEOUT	(5*HZ)
#define	ETHER_TX_TIMEOUT	(5*3*HZ)  // modified by sony

/* added by sony */
typedef enum {
  NOE_NETDEV_TYPE_ETHER,
  NOE_NETDEV_TYPE_WIFI,
  NOE_NETDEV_TYPE_VAP,
} NOE_NETDEV_TYPE;

extern void register_netdevinfo( noe_drv_t *noe_drv, NOE_UINT8 nif_no, NOE_UINT8 nif_subno, NOE_NETDEV_TYPE type, struct net_device *netdev );
extern void unregister_netdevinfo( noe_drv_t *noe_drv, NOE_UINT8 nif_no, NOE_UINT8 nif_subno );
extern bool is_registed_netdevinfo( noe_drv_t *noe_drv, NOE_UINT8 nif_no, NOE_UINT8 nif_subno );
extern struct net_device* netdevinfo2netdev( noe_drv_t *noe_drv, NOE_UINT8 nif_no, NOE_UINT8 nif_subno );

#endif	/* __NOE_DEV_H__ */
