/*
 * usb_drv driver -
 *
 *  Copyright 2009 Sony Corporation.
 *  Copyright (C) Ubiquitous Corp. 2001-2009. All rights reserved.
 *  Copyright (C) 2001-2004 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This driver is based on the 2.6.3 version of drivers/usb/usb-skeleton.c
 * but has been rewritten to be easier to read and use.
 *
 */

#ifndef	__NOE_STD_H__
#define	__NOE_STD_H__

#if	defined(LONG_ISNOT_INT)
typedef	long			NOE_INT32;
typedef	unsigned long		NOE_UINT32;
#else
typedef	int				NOE_INT32;
typedef	unsigned int		NOE_UINT32;
#endif
typedef	short			NOE_INT16;
typedef	unsigned short		NOE_UINT16;
typedef char			NOE_INT8;
typedef	unsigned char		NOE_UINT8;
typedef	NOE_INT8		NOE_CHAR;
typedef	NOE_UINT8		NOE_UCHAR;

#if HAS_LONGLONG == 1
 typedef long long		NOE_INT64;
 typedef unsigned long long	NOE_UINT64;
 #define _LLD "lld"
// #ifndef __LINUX__
//  extern NOE_INT64 strtoll(char *, char **, NOE_INT32);
// #endif
 #define STRTOLL strtoll
#else
 typedef long NOE_INT64;
 typedef unsigned long NOE_UINT64;
 #define _LLD "ld"
 #define STRTOLL strtol
#endif

// we used only lower 7bits for some reason */
typedef	NOE_UINT8	NOE_NIFID;		/* combination of nif_no and nif_subno */

//#if defined(MULTI_NIF)
//#if defined(NOE_BD)
#define	NOE_VERSION_L		((NOE_UINT32)0x20120628L)
#define	NOE_VERSION_H		((NOE_UINT32)0x00000301L)
//#else
//#define	NOE_VERSION_L		((NOE_UINT32)0x20101028L)
//#define	NOE_VERSION_H		((NOE_UINT32)0x00000300L)
//#endif
//#else
//#define	NOE_VERSION_L		((NOE_UINT32)0x20100315L)
//#define	NOE_VERSION_H		((NOE_UINT32)0x00000000L)
//#endif

#define	NOE_FRAME_ALIGN		4
#define	NOE_ROUNDDOWN(x)	((x) & ~(NOE_FRAME_ALIGN - 1))
#define	NOE_ROUNDUP(x)		(((x) + NOE_FRAME_ALIGN - 1) & ~(NOE_FRAME_ALIGN -1))
#define	NOE_MAKEALIGN(x)	(((x) + NOE_FRAME_ALIGN - 1) & ~(NOE_FRAME_ALIGN -1))


// NOE basic frame
typedef struct {
	NOE_UINT32	HostID;			// HOST side ID
	NOE_UINT32	SubSysID;		// SubSystem side ID
	NOE_UINT32	len;			// actual length of payload
	NOE_UINT32	cmd;			// command/response code
	NOE_UINT32	ctrl;
		// extended parameters
	NOE_UINT32	r0;
	NOE_UINT32	r1;
	NOE_UINT32	r2;
#if defined(__UBI_DEBUG) && __UBI_DEBUG
	NOE_UINT32	h_seq;
	NOE_UINT32	s_seq;
	NOE_UINT32	reserve1;
	NOE_UINT32	reserve2;
#endif
} NOE_HEADER;

// defines for header.ctrl
#define	NOE_HC_NIFID	0x0000007f		// NIF ID of this packet
#define	NOE_HC_RSV	0xffffff00		// reserved bit for future used, must be 0

typedef struct {
	NOE_HEADER	h;
	NOE_UINT8	payload[1];
						// payload must be multiple of 4
} NOE_FRAME;

// special SubSysID
#define	NOE_SID_NONE		0		// reserved
#define	NOE_SID_UNBIND		1		// unbind(new) request SubSystem

// special Host ID
#define	NOE_HID_NONE		0		// reserved
#define	NOE_HID_UNBIND		1		// unbind(new) request to HOST

#define	NOE_NUM_OF_DNS	2			// fixed number
typedef struct {
	NOE_UINT32	ip;
	NOE_UINT32	_netmask;               // string "netmask" conflicts with MULTI_NIF
	NOE_UINT32	gateway;
	NOE_UINT32	dns[NOE_NUM_OF_DNS];
	NOE_UINT8	mac[6];
	NOE_UINT8	nif_no;
	NOE_UINT8	reserved[1];			// for padding
        NOE_UCHAR       ifname[16];
 } NOE_NET_INFO;

typedef struct {
	NOE_INT32	desc;
	NOE_UINT32	size;
} NOE_READ_CONTENT;

typedef struct {
	NOE_UINT32	dev_ssid;
	NOE_NIFID	nif_id;
	NOE_UINT8	subcmd;
	NOE_UINT8	mac[6];
	NOE_UINT8	ifname[16];			// terminated by '\0'
} NOE_DEVICE_CHANGED;

// sub-commands
#define	NOE_M_DC_CHANGED		3
#define	   NOE_F_DCC_REMOVED		0
#define	   NOE_F_DCC_ADDED_ETH		1
#define	   NOE_F_DCC_ADDED_WIFI		2
#define	   NOE_F_DCC_ADDED_VAP		3

#define	NOE_MK_NIFID(n, s)	(((s) << 4) | (n))
#define	NOE_NIFID2NO(id)	((id) & 0xf)
#define	NOE_NIFID2SUBNO(id)	(((id) >> 4) & 0x7)
#define	NOE_MAX_NIF_SUBNO	7

struct noe_dlna_highmem_info_st {
    NOE_UINT32     start_add;
    NOE_UINT32     read_size;
    NOE_INT32      last_flag;  //1 is last
    NOE_UINT32     HostID;
    NOE_UINT32     SubSysID;
    NOE_UINT32     actual_read_size;  //return
};

/* added for HDRL_Move */
struct noe_hdrl_request_buffer_info {
    NOE_UCHAR       *p_buffer;         /* highmem unit write start address (from appli) */
    NOE_UINT32      size;              /* highmem write unit size (from appli) */
};

typedef enum {
    HDRL_REPLY_CODE_INVALID,
    HDRL_REPLY_CODE_DATA_COMPLETE,
    HDRL_REPLY_CODE_ACK_RECEIVED,
    HDRL_REPLY_CODE_DATA_FINAL,
} HDRL_REPLY_CODE;


struct _hdrl_received_data {
    NOE_UCHAR       *p_buffer;         /* highmem unit write start address (from appli) */
    NOE_UINT32      size;              /* highmem write unit size (from appli) */
    NOE_UINT32      actual_size;       /* actual size */
};

struct noe_hdrl_received_info {
    HDRL_REPLY_CODE         command;
    union {
        struct _hdrl_received_data  complete_data;
        unsigned long               ack_info; /* dummy */
    };
};

/* added for HDRL_Move */

struct noe_packet_size_status_st {
    NOE_UINT32 mtu_num;
    NOE_UINT32 recv_payload_size;
    NOE_UINT32 send_payload_size;
    NOE_UINT32 recv_packet_size;
    NOE_UINT32 send_packet_size;
};

struct noe_check_flag_data_st {
    unsigned long usb_rec_flag;
    unsigned long net_send_flag;
    unsigned long do_sleep_flag;
    unsigned long hdrl_write_flag;
};


#ifdef NOE_DEF_DUMMY_MODE
struct noe_hdrl_dummy_writeobj_data {
    int size;
    int last_flag;
};
#endif /*NOE_DEF_DUMMY_MODE*/


//
// NOE commands
//
// defines bitwised field for response/command
#define	NOE_M_COMMAND		((NOE_UINT32)0x00ffffffL)
#define	NOE_M_FLAGS		((NOE_UINT32)0xff000000L)
#define	  NOE_F_MORE		((NOE_UINT32)0x01000000L)	// we have more packet
#define	  NOE_F_RESERVE		((NOE_UINT32)0x80000000L)	// reserved to make unsigned status

		/* following bitwised value and flag are combined to len field of NOE_NETWORK_PACKET */
#define NOE_NETWORK_PACKET_OFFLOAD_CHECKSUM		((NOE_UINT32)0x80000000L)
#define	NOE_NETWORK_NIC_NUMBER				((NOE_UINT32)0x7F000000L)		/* network number 0 is first NIC */
#define	NOE_NETWORK_NIC_SHIFT				24

#define	NOE_NETWORK_BASE	0
					/* don't use 0 */
#define	NOE_NETWORK_PACKET					(NOE_NETWORK_BASE + 1)
#define	NOE_NETWORK_STATUS					(NOE_NETWORK_BASE + 2)
#define	NOE_NETWORK_SET_ETHER_PARAMETERS	(NOE_NETWORK_BASE + 3)
#define	NOE_NETWORK_DEVICE_CHANGED		(NOE_NETWORK_BASE + 4)

#define	NOE_INFO_BASE		0x00080
#define	NOE_INFO_GET_VERSION			(NOE_INFO_BASE + 0)

#define	NOE_CONTROL_BASE	0x000c0
#define	NOE_CONTROL_DO_SLEEP			(NOE_CONTROL_BASE + 0)
#define	NOE_CONTROL_DO_DIAGNOSTIC		(NOE_CONTROL_BASE + 1)
#define NOE_CONTROL_MONITOR				(NOE_CONTROL_BASE + 3)
#define NOE_CONTROL_SEND_MAGIC_PACKET_NIF	(NOE_CONTROL_BASE + 4)

#define	NOE_GENERIC_BASE	0x00100
#define	NOE_GENERIC_GET_DATA			(NOE_GENERIC_BASE + 0)
#define	NOE_GENERIC_DONE			(NOE_GENERIC_BASE + 1)

#define	NOE_SOC_BASE		0x00200
#define	NOE_SOC_START_DHCP			(NOE_SOC_BASE + 0)
#define	NOE_SOC_SET_NET_INFO			(NOE_SOC_BASE + 1)
#define	NOE_SOC_GET_NET_INFO			(NOE_SOC_BASE + 2)
#define	NOE_SOC_EXCHG_PAYLOAD_SIZE		(NOE_SOC_BASE + 3)
#define	NOE_SOC_SET_EPHEMERAL_PORT		(NOE_SOC_BASE + 4)
#define	NOE_SOC_CLR_ARPTABLE			(NOE_SOC_BASE + 5)
#define	NOE_SOC_CONFIG_MULTICAST		(NOE_SOC_BASE + 6)
#define	NOE_SOC_NOTIFY_BANDWIDTH		(NOE_SOC_BASE + 7)
#define	NOE_SOC_SET_NET_INFO_NIF		(NOE_SOC_BASE + 8)
#define	NOE_SOC_GET_NET_INFO_NIF		(NOE_SOC_BASE + 9)

#define	NOE_DLNA_BASE		0x00300
#define	NOE_DLNA_START				(NOE_DLNA_BASE + 1)
#define	NOE_DLNA_STOP				(NOE_DLNA_BASE + 2)
#define	NOE_DLNA_UNINIT				(NOE_DLNA_BASE + 3)
#define	NOE_DLNA_SET_PCP_SIZE			(NOE_DLNA_BASE + 7)
#define	NOE_DLNA_GET_DATE			(NOE_DLNA_BASE + 8)
#define	NOE_DLNA_SET_DATE			(NOE_DLNA_BASE + 9)
#define	NOE_DLNA_INIT_NIF			(NOE_DLNA_BASE + 10)
#define	NOE_DLNA_GET_UUID_NIF			(NOE_DLNA_BASE + 11)

#define	NOE_DTCP_BASE		0x00380
#define	NOE_DTCP_UPDATE_SRM			(NOE_DTCP_BASE + 0)
#define	NOE_DTCP_REQ_SAVE_SRM			(NOE_DTCP_BASE + 1)

#define	NOE_DMS_BASE		0x00400
#define	NOE_DMS_READ_CONTENT			(NOE_DMS_BASE + 0)
#define	NOE_DMS_REQ_OPEN_OBJECT			(NOE_DMS_BASE + 3)
#define	NOE_DMS_REQ_READ_OBJECT			(NOE_DMS_BASE + 4)
#define	NOE_DMS_REQ_CLOSE_OBJECT		(NOE_DMS_BASE + 5)
#define	NOE_DMS_SET_STATE_VARIABLE		(NOE_DMS_BASE + 6)
#define	NOE_DMS_GET_STATE_VARIABLE		(NOE_DMS_BASE + 7)
#define	NOE_DMS_REQ_LOCAL_ACTION		(NOE_DMS_BASE + 8)
#define	NOE_DMS_REQ_OPEN_OBJECT_FOR_WRITE	(NOE_DMS_BASE + 9)
#define	NOE_DMS_REQ_WRITE_OBJECT		(NOE_DMS_BASE + 10)
#define	NOE_DMS_REQ_CLOSE_OBJECT_FOR_WRITE	(NOE_DMS_BASE + 11)
#define	NOE_DMS_REQ_CREATE_OBJECT		(NOE_DMS_BASE + 12)
#define	NOE_DMS_REQ_CHANGE_OBJECT_ATTRIBUTE	(NOE_DMS_BASE + 13)
#define	NOE_DMS_REQ_REPORT_EMI			(NOE_DMS_BASE + 15)
#define	NOE_DMS_REQ_MOVE_EVENT			(NOE_DMS_BASE + 16)
#define	NOE_DMS_NOTIFY_BANDWIDTH		(NOE_DMS_BASE + 17)
#define	NOE_DMS_REQ_BROWSE_XML			(NOE_DMS_BASE + 18)
#define	NOE_DMS_SET_STREAM_FILTER		(NOE_DMS_BASE + 19)
#define	NOE_DMS_NOTIFY_STREAM_ERROR		(NOE_DMS_BASE + 20)
#define	NOE_DMS_SET_STREAM_TIMESTAMP_OFFSET	(NOE_DMS_BASE + 21)
#define	NOE_DMS_SET_ACCESS_LIST_MAX		(NOE_DMS_BASE + 22)
#define	NOE_DMS_SET_ACCESS_LIST_MODE		(NOE_DMS_BASE + 23)
#define	NOE_DMS_GET_ACCESS_LIST			(NOE_DMS_BASE + 24)
#define	NOE_DMS_SET_ACCESS_LIST_DEVICE_STATUS	(NOE_DMS_BASE + 25)
#define	NOE_DMS_REQ_PROCESS_HEADER		(NOE_DMS_BASE + 26)
#define	NOE_DMS_STOP_READ	        	(NOE_DMS_BASE + 27)
#define	NOE_DMS_STOP_WRITE		        (NOE_DMS_BASE + 28)
#define	NOE_DMS_DISCARD_POST_CONTEXT		(NOE_DMS_BASE + 29)
#define	NOE_DMS_REQ_PROCESS_HEADER_FOR_WRITE	(NOE_DMS_BASE + 30)
#define NOE_DMS_SET_PCP_UR			(NOE_DMS_BASE + 31)
#define NOE_DMS_GET_PCP_UR			(NOE_DMS_BASE + 32)
#define NOE_DMS_SET_EMI				(NOE_DMS_BASE + 33)
#define NOE_DMS_GET_EMI				(NOE_DMS_BASE + 34)
#define NOE_DMS_REQ_STOP_OBJECT                 (NOE_DMS_BASE + 35)
#define NOE_DMS_REPLACE_DTCP_DESCRIPTOR         (NOE_DMS_BASE + 36)


#define	NOE_DMS_SET_IV				(NOE_DMS_BASE + 97)
#define	NOE_DMS_SET_DATAPORT			(NOE_DMS_BASE + 98)

#if defined(INTERNAL_TESTING) && INTERNAL_TESTING //this is only for testing
#define	NOE_DMS_UPLOAD_LIKE			(NOE_DMS_BASE + 99)
#endif

#define	NOE_DMP_BASE		0x00500
#define	NOE_DMP_BROWSE_DIRECT_CHILDREN		(NOE_DMP_BASE + 0)
#define	NOE_DMP_BROWSE_META_DATA			(NOE_DMP_BASE + 1)
#define	NOE_DMP_BROWSE_XML					(NOE_DMP_BASE + 2)
#define	NOE_DMP_OPEN_OBJECT					(NOE_DMP_BASE + 3)
#define	NOE_DMP_READ_OBJECT					(NOE_DMP_BASE + 4)
#define	NOE_DMP_CLOSE_OBJECT				(NOE_DMP_BASE + 5)
#define NOE_DMP_CREATE_OBJECT				(NOE_DMP_BASE + 6) 
#define NOE_DMP_OPEN_OBJECT_FOR_UPLOAD		(NOE_DMP_BASE + 7) 
#define NOE_DMP_CLOSE_OBJECT_FOR_UPLOAD		(NOE_DMP_BASE + 8)
#define NOE_DMP_UPLOAD_CANCEL				(NOE_DMP_BASE + 9) 
#define NOE_DMP_DO_ACTION					(NOE_DMP_BASE + 10)
#define NOE_DMP_REQUEST_EVENT_FROM_HTTPC	(NOE_DMP_BASE + 11)
#define	NOE_DMP_START_DOWNLOAD				(NOE_DMP_BASE + 12) 
#define	NOE_DMP_STOP_DOWNLOAD				(NOE_DMP_BASE + 13) 
#define	NOE_DMP_START_UPLOAD				(NOE_DMP_BASE + 14) 
#define	NOE_DMP_STOP_UPLOAD					(NOE_DMP_BASE + 15) 
#define NOE_DMP_SET_STREAM_TIMESTAMP_OFFSET (NOE_DMP_BASE + 16)

#define	NOE_UPNPCP_BASE		0x02000
#define	NOE_UPNPCP_GET_DEVICES			(NOE_UPNPCP_BASE + 0)
#define	NOE_UPNPCP_SEND_MSEARCH			(NOE_UPNPCP_BASE + 1)
#define	NOE_UPNPCP_GET_DEVICE_XML		(NOE_UPNPCP_BASE + 2)
#define	NOE_UPNPCP_GET_SERVICE_XML		(NOE_UPNPCP_BASE + 3)
#define NOE_UPNPCP_START_SUBSCRIBE		(NOE_UPNPCP_BASE + 4)
#define NOE_UPNPCP_STOP_SUBSCRIBE		(NOE_UPNPCP_BASE + 5)
#define NOE_UPNPCP_REQ_UPNP_EVENT		(NOE_UPNPCP_BASE + 6)

#define	NOE_USER_BASE		0x0f000
	/* reserved 0x0f000..0x0ffff for user command */

#define	NOE_USER_INVALID_COMMAND			(NOE_USER_BASE + 0xfff)
#define	NOE_USER_RCV_THD_CLOSE_COMMAND			(NOE_USER_BASE + 0x1)


#define	NOE_HOST_IOCTL_BASE	0x10000
#define	NOE_GET_CONTINUOUS			(NOE_HOST_IOCTL_BASE + 0)

//debug settings
#define NOE_MONITOR_DISABLE				(0)
#define NOE_MONITOR_ENABLE				(1 << 0)
#define NOE_MONITOR_TCPIP				(1 << 1)
#define NOE_MONITOR_DTCP				(1 << 2)
#define NOE_MONITOR_HTTP				(1 << 3)
#define NOE_MONITOR_UPNP				(1 << 4)

// custom commands
#define	NOE_PRJ001_FIRST	0x20000
#define	NOE_BDEX_BASE		NOE_PRJ001_FIRST
#define	NOE_BDEX_CREATE_VAP		(NOE_BDEX_BASE + 0)
#define	NOE_BDEX_VAP_IOCTL_GETPARAM	(NOE_BDEX_BASE + 1)
#define	NOE_BDEX_VAP_OPEN		(NOE_BDEX_BASE + 2)
#define	NOE_BDEX_VAP_IOCTL_GIWMODE	(NOE_BDEX_BASE + 3)
#define	NOE_BDEX_VAP_IOCTL_GIWRANGE	(NOE_BDEX_BASE + 4)
#define	NOE_BDEX_VAP_IOCTL_SETPARAM	(NOE_BDEX_BASE + 5)
#define	NOE_BDEX_VAP_IOCTL_SIWESSID	(NOE_BDEX_BASE + 6)
#define	NOE_BDEX_VAP_IOCTL_DELKEY	(NOE_BDEX_BASE + 7)
#define	NOE_BDEX_VAP_IOCTL_IOC_SCAN_REQ	(NOE_BDEX_BASE + 8)
#define	NOE_BDEX_VAP_IOCTL_SCAN_RESULTS	(NOE_BDEX_BASE + 9)
#define	NOE_BDEX_VAP_IOCTL_SETOPTIE	(NOE_BDEX_BASE + 10)
#define	NOE_BDEX_VAP_IOCTL_SETMLME	(NOE_BDEX_BASE + 11)
#define	NOE_BDEX_VAP_IOCTL_GIWFREQ	(NOE_BDEX_BASE + 12)
#define	NOE_BDEX_VAP_IOCTL_GIWAP	(NOE_BDEX_BASE + 13)
#define	NOE_BDEX_VAP_IOCTL_GIWESSID	(NOE_BDEX_BASE + 14)
#define	NOE_BDEX_VAP_IOCTL_SETKEY	(NOE_BDEX_BASE + 15)
#define	NOE_BDEX_DELETE_VAP		(NOE_BDEX_BASE + 16)
#define	NOE_BDEX_VAP_IOCTL_GIWNAME	(NOE_BDEX_BASE + 17)
#define	NOE_BDEX_VAP_IOCTL_GIWENCODE	(NOE_BDEX_BASE + 18)
#define	NOE_BDEX_VAP_IOCTL_GIWRATE	(NOE_BDEX_BASE + 19)
#define	NOE_BDEX_VAP_IOCTL_GIWPOWER	(NOE_BDEX_BASE + 20)
#define	NOE_BDEX_VAP_IOCTL_GIWTXPOW	(NOE_BDEX_BASE + 21)
#define	NOE_BDEX_VAP_IOCTL_GIWFRAG	(NOE_BDEX_BASE + 22)
#define	NOE_BDEX_VAP_IOCTL_GIWRTS	(NOE_BDEX_BASE + 23)
#define	NOE_BDEX_VAP_IOCTL_SIWENCODE	(NOE_BDEX_BASE + 24)
#define	NOE_BDEX_VAP_IOCTL_GETSTATS	(NOE_BDEX_BASE + 25)
#define	NOE_BDEX_IOCTL_GETSTATS		(NOE_BDEX_BASE + 26)
#define	NOE_BDEX_VAP_IOCTL_SIWFREQ	(NOE_BDEX_BASE + 27)
#define	NOE_BDEX_VAP_IOCTL_SIWMODE	(NOE_BDEX_BASE + 28)
#define	NOE_BDEX_VAP_IOCTL_SIWAP	(NOE_BDEX_BASE + 29)
#define	NOE_BDEX_VAP_IOCTL_IWAPLIST	(NOE_BDEX_BASE + 30)
#define	NOE_BDEX_VAP_IOCTL_SIWRATE	(NOE_BDEX_BASE + 31)
#define	NOE_BDEX_VAP_IOCTL_SIWRTS	(NOE_BDEX_BASE + 32)
#define	NOE_BDEX_VAP_IOCTL_SIWFRAG	(NOE_BDEX_BASE + 33)
#define	NOE_BDEX_VAP_IOCTL_SIWTXPOW	(NOE_BDEX_BASE + 34)
#define	NOE_BDEX_VAP_IOCTL_SIWPOWER	(NOE_BDEX_BASE + 35)
#define	NOE_BDEX_VAP_IOCTL_SIWSCAN	(NOE_BDEX_BASE + 36)
#define	NOE_BDEX_WIFI_IOCTL_SETPARAM	(NOE_BDEX_BASE + 37)
#define	NOE_BDEX_WIFI_IOCTL_GETPARAM	(NOE_BDEX_BASE + 38)
#define	NOE_BDEX_WIFI_IOCTL_SETCOUNTRY	(NOE_BDEX_BASE + 39)
#define	NOE_BDEX_WIFI_IOCTL_GETCOUNTRY	(NOE_BDEX_BASE + 40)
#define	NOE_BDEX_WIFI_OPEN		(NOE_BDEX_BASE + 41)
#define	NOE_BDEX_WIFI_STOP		(NOE_BDEX_BASE + 42)
#define	NOE_BDEX_VAP_STOP		(NOE_BDEX_BASE + 43)
#define	NOE_BDEX_WIFI_TX_TIMEOUT	(NOE_BDEX_BASE + 44)
#define	NOE_BDEX_VAP_IOCTL_GETCHANINFO	(NOE_BDEX_BASE + 45)
#define	NOE_BDEX_VAP_IOCTL_GETCHANLIST	(NOE_BDEX_BASE + 46)
#define	NOE_BDEX_VAP_IOCTL_GETPARAM_STAINFO		(NOE_BDEX_BASE + 47)
#define	NOE_BDEX_VAP_IOCTL_IOC_CANCEL_SCAN		(NOE_BDEX_BASE + 48)
//P2P>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
#define	NOE_BDEX_VAP_IOCTL_IOC_P2P_FETCH_FRAME		(NOE_BDEX_BASE + 80)
#define	NOE_BDEX_VAP_IOCTL_APPIE_FRAME_BEACON		(NOE_BDEX_BASE + 81)
#define	NOE_BDEX_VAP_IOCTL_APPIE_FRAME_PROBE_RESP	(NOE_BDEX_BASE + 82)
#define	NOE_BDEX_VAP_IOCTL_IOC_P2P_SET_CHANNEL		(NOE_BDEX_BASE + 83)
#define	NOE_BDEX_VAP_IOCTL_IOC_P2P_SEND_ACTION		(NOE_BDEX_BASE + 84)
#define	NOE_BDEX_VAP_IOCTL_IOC_P2P_FIND_BEST_CHANNEL	(NOE_BDEX_BASE + 85)
//P2P<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

#define	NOE_BDEX_NETIF_BASE		(NOE_PRJ001_FIRST+0x100)
#define	NOE_BDEX_NETIF_SET_CARRIER			(NOE_BDEX_NETIF_BASE + 0)
#define	NOE_BDEX_NETIF_UPDATE_FLAGS_S2H			(NOE_BDEX_NETIF_BASE + 1)
#define	NOE_BDEX_NETIF_UPDATE_FLAGS_H2S			(NOE_BDEX_NETIF_BASE + 2)
#define	NOE_BDEX_NETIF_WIRELESS_SEND_EVENT		(NOE_BDEX_NETIF_BASE + 3)
#define	NOE_BDEX_NETIF_WIRELESS_SEND_EVENT_SIOCGIWAP	(NOE_BDEX_NETIF_BASE + 4)
#define	NOE_BDEX_NETIF_UPDATE_STATS			(NOE_BDEX_NETIF_BASE + 5)
#define	NOE_BDEX_NETIF_UPDATE_IW_STATS			(NOE_BDEX_NETIF_BASE + 6)
//P2P>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
#define	NOE_BDEX_NETIF_WIRELESS_SEND_EVENT_UINT32	(NOE_BDEX_NETIF_BASE + 10)
//P2P<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

#define	NOE_PRJ001_LAST		0x20fff

//
//
// Sleep mode definitions
//
//
#define NOE_SLEEP_MODE_SLEEP	(0)
#define NOE_SLEEP_MODE_STANDBY	(1)

//
//
// Band width control definitions
//
//
#define	NOE_BWC_OUTBOUND		0		// from subsystem to somewhere
#define	NOE_BWC_INBOUND			1		// from somewhere to subsystem

//
//
// public definition for DLNA application
//
//
#include "dlna_api.h"

#endif	//__NOE_STD_H__
