/*
** NOE 2.1 Linux kernel driver
** driver, usb part
**
** Copyright (C) Ubiquitous Corp. 2001-2010. All rights reserved.
**
** $Revision: #1 $	$Date: 2012/08/03 $
**
*/
//  original driver's comments.
/*
 * usb_drv driver -
 *
 * Copyright (C) 2001-2004 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 * This driver is based on the 2.6.3 version of drivers/usb/usb-skeleton.c
 * but has been rewritten to be easier to read and use.
 *
 * for NOE: use skeleton.c [2007-08-24]
 */
/*
 * network driver -
 *
 *  Copyright (c) 2002 Petko Manolov (petkan@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * for NOE: use from [2007-10-24]
 */

/*
** We use minimum lock region for speed. 
** we don't protect open/delete, each i/o
*/

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/mutex.h>
#include <linux/sockios.h>
#include <linux/stat.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/interrupt.h>
#include <linux/wireless.h>
#include <linux/rtnetlink.h>
#include <linux/kthread.h>
#include <net/iw_handler.h>
#include <linux/proc_fs.h> /* added by sony */

#include "noe_std.h"
#include "noe_dev.h"
#include "noe_driver_if.h"
#include "noe_drv_config.h"

#include "noe_kdef.h"

/* to prevent a race between open and disconnect */
static DEFINE_MUTEX(noe_open_lock);

/* Get a minor range for your devices from the usb maintainer */
#define NOE_USB_MINOR_BASE	192

/* NOE Function */
#define NOE_BUFSIZE		HOST_DRIVER_MAXPACKET //Max NOE buf

const char driver_name[] = "noe_drv";
const char driver_version[] = "v0.2.9990";

static struct usb_driver noe_driver;
static void noe_delete(struct kref *kref);

/********************** added defines by sony ******************/
typedef struct {
    struct list_head   list;
    noe_rcv_usb_data  *data;
} noe_rcv_thd_buffer_t;

typedef enum {
    NOE_HDRL_COMMAND_WRITE_DATA,
    NOE_HDRL_COMMAND_FLUSH_DATA,
    NOE_HDRL_COMMAND_TERMINATE,
} NOE_HDRL_COMMAND;

typedef struct {
    struct list_head   list;
    NOE_HDRL_COMMAND   command;
    noe_rcv_usb_data  *data;
} noe_hdrl_data_t;

/* added for HDRL_Move */
typedef struct {
    struct list_head                list;
    struct noe_hdrl_received_info   received_info;
} noe_write_unit_data_t;
/* */

#define DEF_NFE_WAIT_SLEEP_TIMEOUT    HZ

static unsigned long g_noe_usb_status = 0;
enum{
    DEF_NFE_USB_STATUS_INIT,
};

static int g_noe_usb_mtu_num = NFE_DEFAULT_MULTI_PACKET_CNT;
module_param( g_noe_usb_mtu_num, int, S_IRUGO );

#if defined RELEASE_VERSION_ONLY
static bool g_noe_debug_console = false;
#else
static bool g_noe_debug_console = true;
#endif

/********************** added functions by sony ******************/
static inline bool _is_noe_rcv_thd( NOE_UINT32 HostID );
static int _get_noe_rcv_thd_data( noe_drv_t *noe_drv, void *user_buffer, NOE_pFRAME *noe_packet );
static int _read_noe_data( noe_drv_t *noe_drv, NOE_UINT32 HostID, void* user_header, void *user_payload );
static int _send_noe_rcv_thd_data( noe_drv_t *noe_drv, noe_rcv_usb_data *data );

static void _noe_pre_close( noe_drv_t *noe_drv );
static void _noe_init( noe_drv_t *noe_drv );

static int _start_read_object( noe_drv_t *noe_drv, int *highmem_total_send_len );

static bool _is_write_object_cmd( noe_drv_t *noe_drv, NOE_HEADER* buf );
static int _put_write_object_cmd( hdrl_write_queue *queue, NOE_HDRL_COMMAND cmd, noe_rcv_usb_data* data );
static struct task_struct* _create_write_object_thread( noe_drv_t *noe_drv );
static void _destroy_write_object_thread( noe_drv_t *noe_drv );

static NOE_UINT32 _get_uuid_nif( noe_drv_t *noe_drv, NOE_INT32 DeviceID, NOE_UCHAR *uuid, NOE_UINT32 index );
static NOE_UINT32 _get_net_info_nif( noe_drv_t *noe_drv, NOE_NET_INFO *netinfo, NOE_UINT32 index );
static NOE_UINT32 _set_net_info_nif( noe_drv_t *noe_drv, NOE_UINT32 num_netinfo, NOE_NET_INFO *netinfo );
//static NOE_UINT32 _send_magic_packet_nif( noe_drv_t *noe_drv, NOE_UCHAR *mac, NOE_UINT32 index );

/* added for HDRL_Move */
static int _enqueue_hdrl_write_unit( request_write_queue *queue_info, void* buffer, int size );
static int _fill_hdrl_write_unit( request_write_queue *queue_info, void* data, int size, bool ack );
static int _flush_hdrl_write_unit( request_write_queue *queue_info );
static void _revoke_hdrl_request_write_queue( noe_drv_t *noe_drv );
static int  _wait_hdrl_queue_ready( request_write_queue *queue_info );
static int  _wait_hdrl_unit_filled( request_write_queue *queue_info, struct noe_hdrl_received_info *info );

static void _setup_procfs_nif( noe_drv_t *noe_drv );
static void _terminate_procfs_nif( void );
static void _setup_procfs( noe_drv_t *noe_drv );
static void _terminate_procfs( noe_drv_t *noe_drv );
static void _start_network_monitor_thread( noe_drv_t *noe_drv );
static void _stop_network_monitor_thread( noe_drv_t *noe_drv );
static void _noe_data_dump( noe_drv_t *noe_drv, int pos, NOE_UINT32 HostID );

void net_drv_free_rx_buf( noe_drv_t *noe_drv, noe_rcv_usb_data *data ); /* net_drv.c */

/********************** debug control by sony ******************/
#define NOE_DEBUG_ERROR
#define NOE_DEBUG_WARNING
#define NOE_DEBUG_INFO
#undef  NOE_DEBUG_DP
#undef  NOE_DEBUG_HDRL_SVR
#undef  NOE_DEBUG_DLNA_SVR
#undef  NOE_DEBUG_TRACE_BULK_TRANSFER
#undef  NOE_DEBUG_TRACE_IOCTL
#include "noe_debug.h"
#include "noe_command_debug.h"

/* added by sony */
static struct list_head           g_netdevinfo_list[NOE_DEVICE_TYPE_MAX];
static wait_queue_head_t          g_rmmod_ready_wq;
/*#define REMOVE_DRIVER_READY_TIMEOUT   5*HZ*/
#define REMOVE_DRIVER_READY_TIMEOUT   30*HZ

static void _init_netdev_info( noe_drv_t *noe_drv )
{
    int i,j;

    for( i = 0; i < MAX_NIF; i++ ){
        for( j = 0; j < MAX_NIF_SUB; j++ ){
            noe_drv->netdev_info[i][j].netdev = NULL;
        }
    }
}

void register_netdevinfo( noe_drv_t *noe_drv, NOE_UINT8 nif_no, NOE_UINT8 nif_subno, NOE_NETDEV_TYPE type, struct net_device *netdev )
{
    struct netdev_info *netdev_info;

    NOE_INFO("Register NetDev#%d[%d][%d] %s\n", type, nif_no, nif_subno, netdev->name);

    netdev_info = &noe_drv->netdev_info[nif_no][nif_subno];
    netdev_info->netdev = netdev;
    netdev_info->type = type;
    netdev_info->nif = nif_no;
    netdev_info->nif_sub = nif_subno;

    list_add_tail( &netdev_info->list, &g_netdevinfo_list[type] );
}

void unregister_netdevinfo( noe_drv_t *noe_drv, NOE_UINT8 nif_no, NOE_UINT8 nif_subno )
{
    struct netdev_info *netdev_info;

    netdev_info = &noe_drv->netdev_info[nif_no][nif_subno];

    NOE_INFO("Unregister NetDev#%d[%d][%d] %s\n", netdev_info->type, netdev_info->nif, netdev_info->nif_sub, netdev_info->netdev->name);

    netdev_info->netdev = NULL;
    netdev_info->type = NOE_DEVICE_TYPE_UNKNOWN;

    list_del_init( &netdev_info->list );

    if(waitqueue_active(&g_rmmod_ready_wq)){
        wake_up(&g_rmmod_ready_wq);
    }
}

struct net_device* netdevinfo2netdev( noe_drv_t *noe_drv, NOE_UINT8 nif_no, NOE_UINT8 nif_subno )
{
    return noe_drv->netdev_info[nif_no][nif_subno].netdev;
}

bool is_registed_netdevinfo( noe_drv_t *noe_drv, NOE_UINT8 nif_no, NOE_UINT8 nif_subno )
{
    return (netdevinfo2netdev( noe_drv, nif_no, nif_subno ) != NULL);
}

static int _wait_rmmod_ready(void)
{
    /* All WiFi devices are removed */
    if( wait_event_timeout( g_rmmod_ready_wq, list_empty(&g_netdevinfo_list[NOE_DEVICE_TYPE_WIFI]), REMOVE_DRIVER_READY_TIMEOUT ) == 0 ){
        NOE_ERROR("waiting for completion to NOE rmmod ready is timeout!!\n");
        return -1;
    }

    return 0;
}

/* changed by sony */
static noe_packet_buffer_t *
_find_packet( noe_packet_buffer_list_t *buffer_list, NOE_UINT32 HostID )
{
        noe_packet_buffer_t *c;
        struct list_head *used_list;

        used_list = &buffer_list->used_list;

        list_for_each_entry( c, used_list, list ){
		if(c->HostID == HostID)
			return(c);
        }

	return(NULL);
}

#if 0
//for small code
#define	get_free_packet(pb)	find_packet((pb), NOE_HID_NONE)
#else
/* changed by sony */
static noe_packet_buffer_t *
_get_free_packet( noe_packet_buffer_list_t *buffer_list )
{
        noe_packet_buffer_t *c;
        struct list_head *free_list, *used_list;

        free_list = &buffer_list->free_list;
        used_list = &buffer_list->used_list;

        if( list_empty(free_list) ){
                return NULL;
        }

        c = list_first_entry( free_list, noe_packet_buffer_t, list );
        list_del_init(&c->list);
        list_add_tail( &c->list, used_list );

#ifdef DEBUG_PACKET_BUFFER_LIST
        buffer_list->free_cnt--;
        buffer_list->used_cnt++;
        printk("### PACKET BUFFER(allc) %02d/%02d ####\n", buffer_list->free_cnt, buffer_list->used_cnt);
#endif

        return c;
}
#endif

/* added by sony */
static void
_release_used_packet( noe_drv_t *noe_drv, noe_packet_buffer_t *used )
{
        noe_packet_buffer_t *c,*n;
        struct list_head *free_list, *used_list;
        bool found = false;
        noe_packet_buffer_list_t *buffer_list = &noe_drv->packet_buffer_by_hostid;

        free_list = &buffer_list->free_list;
        used_list = &buffer_list->used_list;

        list_for_each_entry_safe( c, n, used_list, list ){
                if( c == used ){
                    list_del_init(&c->list);
                    found = true;
                    break;
                }
        }

        if(!found){
            NOE_ERROR("cannot found used packet data\n");
            return;
        }

        net_drv_free_rx_buf( noe_drv, c->data );
        c->data = NULL;
	c->HostID = NOE_HID_NONE;

        list_add_tail( &c->list, free_list );
#ifdef DEBUG_PACKET_BUFFER_LIST
        buffer_list->used_cnt--;
        buffer_list->free_cnt++;
        printk("### PACKET BUFFER(free) %02d/%02d ####\n", buffer_list->free_cnt, buffer_list->used_cnt);
#endif
}

/* added by sony */
static void
_release_all_packet( noe_drv_t *noe_drv )
{
        noe_packet_buffer_t *c,*n;
        struct list_head *free_list, *used_list;
        noe_packet_buffer_list_t *buffer_list = &noe_drv->packet_buffer_by_hostid;

        free_list = &buffer_list->free_list;
        used_list = &buffer_list->used_list;

        list_for_each_entry_safe( c, n, used_list, list ){
                list_del_init(&c->list);

                net_drv_free_rx_buf( noe_drv, c->data );
                c->data = NULL;
                c->HostID = NOE_HID_NONE;

                list_add_tail( &c->list, free_list );
#ifdef DEBUG_PACKET_BUFFER_LIST
                buffer_list->used_cnt--;
                buffer_list->free_cnt++;
#endif
        }
#ifdef DEBUG_PACKET_BUFFER_LIST
        printk("### PACKET BUFFER(all free) %02d/%02d ####\n", buffer_list->free_cnt, buffer_list->used_cnt);
#endif
}

static noe_wait_queue_t *
find_wait_queue(noe_wait_queue_t *wq, NOE_UINT32 HostID)
{
	int	i;

	for(i = 0;i < WQ_RINGS;i++) {
		if(wq->HostID == HostID)
			return(wq);
		wq++;
	}
	return(NULL);
}

#if 0
//for small code
#define	get_free_wait_queue(wq)	find_wait_queue((wq), NOE_HID_NONE)
#else
static noe_wait_queue_t *
get_free_wait_queue(noe_wait_queue_t *wq)
{
	int	i;

	for(i = 0;i < WQ_RINGS;i++) {
		if(wq->HostID == NOE_HID_NONE)
			return(wq);
		wq++;
	}
	return(NULL);
}
#endif

	/*
	** This routine is called within irq context,
	** therefore we cannot copy data to user space.
	*/
void
noe_rec_data(struct urb *urb, noe_drv_t *noe_drv, noe_rcv_usb_data *data) /* modified API by sony */
{
	noe_wait_queue_t	*wq;
	noe_packet_buffer_t	*pb;
	unsigned long		flags;
        NOE_UINT32              HostID;
        NOE_pFRAME             *frame;

        frame = GET_RCV_USB_DATA_BODY(data);

#ifdef NOE_DEF_CMD_TRACE
        _trace_rcv_cmd(dev, payload); /* added by sony */
#endif

        HostID = frame->h.HostID;

        PRINT_PACKET("RECV", &(frame->h));
        DUMP_USB_PACKET( urb->actual_length, frame );

        /* added by sony */
        if( test_bit(DEF_NFE_DEBUG_NETWORK_MONITOR, &noe_drv->debug_status) ){
            noe_drv->debug_dlna_rcv_len += urb->actual_length;
        }

        /* added by sony */
        if( test_bit(DEF_NFE_SLEEP_MODE, &noe_drv->do_sleep_flag) ){
            if( HostID == noe_drv->do_sleep_hostid ){
                NOE_INFO("set RECEIVE_DO_SLEEP_BIT HostID=%x\n", (unsigned int)HostID);
                set_bit( DEF_NFE_SLEEP_RECEIVE_PACKET, &noe_drv->do_sleep_flag);
		if( waitqueue_active(&noe_drv->do_sleep_wait_queue) ){
                  wake_up_interruptible(&noe_drv->do_sleep_wait_queue);
                }
            }
        }

        /* added by sony */
        if( _is_write_object_cmd(noe_drv, &frame->h) ){
            SET_RCV_USB_DATA_TYPE(data, NFE_RX_DATA_TYPE_WRITE_OBJECT);
            _put_write_object_cmd( &noe_drv->hdrl_write_queue, NOE_HDRL_COMMAND_WRITE_DATA, data );
            /* free buffer later in write_object_thread */
            return;
        }

        if( _is_noe_rcv_thd(HostID) ){
            /* modified by sony */
            SET_RCV_USB_DATA_TYPE(data, NFE_RX_DATA_TYPE_RCV_THD_DATA);
            _send_noe_rcv_thd_data(noe_drv, data);
            /* free buffer later in user ioctl context */
        }
        else{
            spin_lock_irqsave(&noe_drv->rd_ioctl_lock, flags);
            pb = _get_free_packet(&noe_drv->packet_buffer_by_hostid);
            if(pb) {
                SET_RCV_USB_DATA_TYPE(data, NFE_RX_DATA_TYPE_OTHER);
                pb->data = data; /* modified by sony */
		pb->HostID = HostID; /* modified by sony */
		pb->length = urb->actual_length;

		wq = find_wait_queue(noe_drv->wait_queue, HostID);
		if(wq) {
			wq->HostID = NOE_HID_NONE;
			wake_up_interruptible(&wq->ioctl_wait);
		}
            }
            else {
                NOE_ERROR("no more packet buffer\n");
                net_drv_free_rx_buf(noe_drv, data);
            }
            spin_unlock_irqrestore(&noe_drv->rd_ioctl_lock, flags);
            /* free buffer later in user ioctl context */
        }
}

#define NOE_OPEN_RETRY_CNT 5 /* add by sony */
static int
noe_open(struct inode *inode, struct file *file)
{
	int			subminor;
	noe_drv_t		*dev;
	struct usb_interface	*interface;
        int                     i;

	subminor = iminor(inode);

        for( i = 0; i< NOE_OPEN_RETRY_CNT; i++ ){ /* modified by sony */
            interface = usb_find_interface(&noe_driver, subminor);
            if( interface != NULL ) break;
            set_current_state(TASK_INTERRUPTIBLE);
            if( schedule_timeout(HZ) != 0 ){
              return -ERESTARTSYS;
            }
        }

	if(interface == NULL) {
		NOE_ERROR ("%s - error, can't find device for minor %d\n", __func__, subminor);
		return(-ENODEV);
	}

	dev = usb_get_intfdata(interface);
	if(dev == NULL)
		return(-ENODEV);

	if(test_and_set_bit(DRIVER_OPENED, &dev->flags))
		return(-EBUSY);

	/* increment our usage count for the device */
	kref_get(&dev->kref);

	/* save our object in the file's private structure */
	file->private_data = dev;

	return(0);
}

static int
noe_release(struct inode *inode, struct file *file)
{
	noe_drv_t	*dev;
	int		i;
        unsigned long   flags;

	dev = (noe_drv_t *)file->private_data;
	if(dev == NULL)
		return(-ENODEV);

	for(i = 0;i < WQ_RINGS;i++)
		dev->wait_queue[i].HostID = NOE_HID_NONE;

        /* modified by sony */
        spin_lock_irqsave(&dev->rd_ioctl_lock, flags);
        _release_all_packet(dev);
        spin_unlock_irqrestore(&dev->rd_ioctl_lock, flags);

	clear_bit(DRIVER_OPENED, &dev->flags);

	kref_put(&dev->kref, noe_delete);

	return(0);
}

static ssize_t
noe_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	return(0);
}

static ssize_t
noe_write(struct file *file, const char *user_buffer, size_t count, loff_t *ppos)
{
	return(0);
}

static void
send_noe_callback(struct urb *urb)
{
	noe_drv_t	*dev;

	dev = urb->context;
	if(dev == NULL) {
		NOE_WARNING("send_noe_callback no dev\n");
		return;
	}
	if(urb->status) {
		NOE_INFO("send_noe status %d\n", urb->status);
	}

        clear_bit( DEF_NFE_SEND_NOE_DATA, &dev->noe_send_status ); /* modified by sony */
	wake_up_interruptible(&dev->ioctl_noe);
}

static int
__noe_send_data(noe_drv_t *dev, int *actual_length, unsigned char *buf, long pkt_len)
{
	int	retval;
	int     ret = 0;
   
	usb_fill_bulk_urb(dev->write_urb, dev->udev, usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr), buf, pkt_len, send_noe_callback, dev);

        NOE_TRACE_BULK("send to write_urb len=%ld\n", pkt_len);

	/* to support zero length packet */
	dev->write_urb->transfer_flags |= URB_ZERO_PACKET;

        set_bit(DEF_NFE_SEND_NOE_DATA, &dev->noe_send_status); /* modified by sony */

	if((retval = usb_submit_urb(dev->write_urb, GFP_ATOMIC)) == 0) {
            /* modified by sony */
            retval = wait_event_interruptible_timeout( dev->ioctl_noe,
                                                       !test_bit(DEF_NFE_SEND_NOE_DATA, &dev->noe_send_status),
                                                       (10*HZ) );
            if( retval == 0 ){
                NOE_ERROR("%s: Send NOE Data timeout\n", __func__);
                ret =  -EBUSY;
		goto _path_send_data_done;
            }
            else if( retval < 0 ){
                ret = -ERESTARTSYS;
		goto _path_send_data_done;
            }
            else{
                *actual_length = dev->write_urb->actual_length;
                NOE_TRACE_BULK("send to write_urb done(%d) \n", *actual_length);
            }
	}
	else {
            /* some error occurred, upper layer will handle it */
            /* modified by sony */
            NOE_ERROR("failed to send to write_urb(%d) \n", retval);
            *actual_length = 0;
            clear_bit( DEF_NFE_SEND_NOE_DATA, &dev->noe_send_status ); 
	}

 _path_send_data_done:
        return(ret);
}

/* added by sony */
static int
_noe_send_data( noe_drv_t *dev, unsigned char *buf, long pkt_len )
{
    int	actual_length = 0;
    int retval = 0;
    unsigned char *p = buf;

    if( test_bit(DEF_NFE_DEBUG_NETWORK_MONITOR, &dev->debug_status) ){
        dev->debug_noe_data_snd_len += pkt_len;
    }

    PRINT_PACKET("SEND", &((NOE_pFRAME *)buf)->h);
    DUMP_USB_PACKET( pkt_len, buf );

    ENTRY_CMD_HISTORY( dev, DEBUG_HISTORY_TYPE_SEND, ((NOE_pFRAME *)buf)->h.cmd, ((NOE_pFRAME *)buf)->h.len );

    while(1) {
        NOE_DP("execute %s len=%d\n", __func__, (int)pkt_len);
        retval = __noe_send_data(dev, &actual_length, p, pkt_len);
        if(retval) {
            NOE_ERROR("%s: cannot noe_send_data %d\n", __func__, retval);
            break;
        }
        p += actual_length;
        pkt_len -= actual_length;
        if(pkt_len <= 0)
          break;
    }

    return retval;
}

/* added by sony */
static int
noe_send_data( noe_drv_t *dev, unsigned char *buf, long pkt_len )
{
    int retval;
    struct net_device *netdev;

    netdev = get_current_netdev(dev);
    if( netdev ) netif_stop_queue(netdev);

    dev->latest_noe_cmd = ((NOE_HEADER *)buf)->cmd;

    if( wait_event_interruptible( dev->bulkout_wait, !test_and_set_bit(RIGHT_TO_USE_BULKOUT, &dev->flags) ) < 0 ){
        if( netdev ) netif_wake_queue(netdev);
        return -ERESTARTSYS;
    }

    retval = _noe_send_data( dev, buf, pkt_len );

    clear_bit(RIGHT_TO_USE_BULKOUT, &dev->flags);
    if(waitqueue_active( &dev->bulkout_wait )){
        wake_up_interruptible( &dev->bulkout_wait );
    }

    if( netdev) netif_wake_queue(netdev);

    return retval;
}

static int
NOEki_send_packet(noe_drv_t *dev, int pkt_len)
{
    /* modified by sony */
    return noe_send_data(dev, dev->bulk_out_buffer, pkt_len);
}

void
NOEki_send_noe_header_payload(noe_drv_t *dev, NOE_HEADER *h, void *payload)
{
	NOE_HEADER	*bh;

	mutex_lock(&dev->write_mutex);
	bh = (NOE_HEADER *)dev->bulk_out_buffer;
	bh->SubSysID = host2sub32(h->SubSysID);
	bh->HostID = host2sub32(h->HostID);
	bh->cmd = host2sub32(h->cmd);
	bh->len = host2sub32(h->len);
	bh->ctrl = host2sub32(h->ctrl);
	bh->r0 = h->r0;
	bh->r1 = h->r1;
	bh->r2 = h->r2;
	WARN_ON(h->len > NOE_BUFSIZE);
	if(h->len)
		memcpy(bh + 1, payload, h->len);

	NOEki_send_packet(dev, NOE_MAKEALIGN(h->len) + sizeof(NOE_HEADER));
	mutex_unlock(&dev->write_mutex);
}

int
NOEki_wait_normal_packet(noe_drv_t *dev, NOE_HEADER *h, void *payload)
{

	NOE_UINT32		seHostID;	/* Subsystem's endian HostID */
	unsigned long		flags;
	noe_wait_queue_t	*wq;
	noe_packet_buffer_t	*pb;
	int			retval;
	long			timeout;
        NOE_pFRAME             *rcv_frame;

	seHostID = host2sub32(h->HostID);
	timeout = (h->HostID == NOE_HID_UNBIND) ? 10 * HZ : HZ;
	wq = NULL;
	while(1) {
		spin_lock_irqsave(&dev->rd_ioctl_lock, flags);
		pb = _find_packet(&dev->packet_buffer_by_hostid, seHostID);
		if(pb)
			break;
		if(wq == NULL) {
		wq = get_free_wait_queue(dev->wait_queue);
			if(wq == NULL) {
				spin_unlock_irqrestore(&dev->rd_ioctl_lock, flags);
				NOE_ERROR("no more wait queue\n");
				return(-EBUSY);
			}
			wq->HostID = seHostID;
		}
		spin_unlock_irqrestore(&dev->rd_ioctl_lock, flags);
		retval = wait_event_interruptible_timeout(wq->ioctl_wait, (wq->HostID == NOE_HID_NONE), timeout);
		if(retval < 0)
			return(retval);
	}
	spin_unlock_irqrestore(&dev->rd_ioctl_lock, flags);
        /* modified by sony */
        rcv_frame = GET_RCV_USB_DATA_BODY(pb->data);
	h->SubSysID = sub2host32(rcv_frame->h.SubSysID);
	h->HostID = sub2host32(rcv_frame->h.HostID);
	h->cmd = sub2host32(rcv_frame->h.cmd);
	h->len = sub2host32(rcv_frame->h.len);
	h->ctrl = sub2host32(rcv_frame->h.ctrl);
		/* caller must care about extra register endians */
	h->r0 = rcv_frame->h.r0;
	h->r1 = rcv_frame->h.r1;
	h->r2 = rcv_frame->h.r2;
	if(h->len && payload)
            memcpy(payload, (void*)rcv_frame + sizeof(NOE_HEADER), h->len);

        /* added by sony */
	spin_lock_irqsave(&dev->rd_ioctl_lock, flags);
        _release_used_packet( dev, pb );
	spin_unlock_irqrestore(&dev->rd_ioctl_lock, flags);

	return(0);
}

#if 0
NOE_INT32
NOE_GetVersion(noe_drv_t *dev)
{
	NOE_UINT32	version[2];
	NOE_HEADER	h;

	h.SubSysID = NOE_SID_UNBIND;
	h.HostID = 123;
	h.len = 0;
	h.cmd = NOE_INFO_GET_VERSION;
	h.ctrl = noe_priv->nif_id;

	NOEki_send_noe_header_payload(dev, &h, NULL);

	NOEki_wait_normal_packet(dev, &h, version);

	if(h.cmd == NOE_ACK) {
		printk("NOE Firmware Version %08x:%08x\n", sub2host32(version[1]), sub2host32(version[0]));
		return(NOE_E_OK);
	}

	return(NOE_ERR_DEV2API(h.cmd));
}
#endif

static int
noe_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	unsigned char	*user_buffer = (unsigned char *)arg;
	int		retval = 0;
	unsigned char	index;
	noe_drv_t	*noe_drv;
	NOE_pFRAME	*hp;
	NOE_pFRAME	noe_packet;
	unsigned long	len;
	unsigned char	num_nif;
	long		pkt_len;
        int             res;

	noe_drv = (noe_drv_t *)file->private_data;

	switch(cmd) {
	case NOE_WRITE_IOCTL: /* noe write function */
		/* one by one */
		mutex_lock(&noe_drv->write_mutex);

		if(copy_from_user(noe_drv->bulk_out_buffer, user_buffer, sizeof(NOE_pFRAME))) {
			retval = -EACCES;
			goto _write_error;
		}
#ifndef DEBUG_NOE_COMMAND
                NOE_TRACE_IOCTL("call ioctl:NOE_WRITE_IOCTL<%08x>(cmd:0x%05x/len:%d)\n",
                                ((NOE_pFRAME*)noe_drv->bulk_out_buffer)->h.HostID,
                                ((NOE_pFRAME*)noe_drv->bulk_out_buffer)->h.cmd,
                                ((NOE_pFRAME*)noe_drv->bulk_out_buffer)->h.len);
#endif
		hp = (NOE_pFRAME *)noe_drv->bulk_out_buffer;
#if defined(__UBI_DEBUG) && __UBI_DEBUG
		hp->h.h_seq = host2sub32(++noe_host_seq);
#endif
		//info("write ioctl: hostid=%x cmd=%x len=%d", sub2host32(hp->h.HostID), sub2host32(hp->h.cmd), sub2host32(hp->h.len));
				// packet's byte order is equal to subsystem's.
		len = sub2host32p(&(hp->h.len));
		if(len > HOST_DRIVER_MAXPAYLOAD) {
			NOE_ERROR("large:%d\n",(int)len);
			len = HOST_DRIVER_MAXPAYLOAD;
		}
		if(len && copy_from_user((void*)noe_drv->bulk_out_buffer + sizeof(NOE_HEADER), hp->payload, len)) {
			retval = -EACCES;
			goto _write_error;
		}
		pkt_len = NOE_MAKEALIGN(len) + sizeof(NOE_HEADER);

                /* added by sony */
                if( hp->h.cmd == NOE_CONTROL_DO_SLEEP ){
                  noe_drv->do_sleep_hostid = hp->h.HostID;
                }

                retval = noe_send_data(noe_drv, noe_drv->bulk_out_buffer, pkt_len); /* modified by sony */

	_write_error:;
		mutex_unlock(&noe_drv->write_mutex);
		break;

	case NOE_READ_IOCTL:
		if(copy_from_user(&noe_packet, user_buffer, sizeof(NOE_pFRAME)))
			return(-EACCES);
		//info("read ioctl: hostid=%x", sub2host32(noe_packet.h.HostID));
                /************ modified by sony ***********/
                if( _is_noe_rcv_thd( noe_packet.h.HostID ) ){
                  /*Get rev thd data and copy data to user buffer*/
                  NOE_TRACE_IOCTL("call ioctl:NOE_READ__IOCTL<rcv_thd_data>\n");
                  res = _get_noe_rcv_thd_data( noe_drv, user_buffer, &noe_packet );
                } else {
                  /*Get rev data and copy data to user buffer*/
#ifndef DEBUG_NOE_COMMAND
                  NOE_TRACE_IOCTL("call ioctl:NOE_READ__IOCTL<%08x>\n", noe_packet.h.HostID);
#endif
                  res = _read_noe_data( noe_drv, noe_packet.h.HostID, user_buffer, noe_packet.payload );
                }

		if(res)
			return(-EACCES);
		break;

	case NOE_GET_NET_NAME: /* get ether name */
        {
                struct net_device *netdev;
                NOE_TRACE_IOCTL("call ioctl:NOE_GET_NET_NAME\n");
		if(copy_from_user(&index, user_buffer, 1))
			return(-EACCES);

			/* slot #0 is primary device for this NIF */
		netdev = netdevinfo2netdev( noe_drv, index, 0 );
		if(netdev) {
			if(copy_to_user(user_buffer, netdev->name, strlen(netdev->name) + 1))
				return(-EACCES);
			return(0);
		}
		return(-EINVAL);
        }
	case NOE_GET_NUM_NIF:
                NOE_TRACE_IOCTL("call ioctl:NOE_GET_NUM_NIF\n");
		num_nif = (unsigned char)noe_drv->num_nif;
		if(copy_to_user(user_buffer, &num_nif, sizeof(num_nif)))
			return(-EACCES);
		break;

	case NOE_DO_SLEEP_MODE: /* enter do_sleep_mode */
        {
                NOE_TRACE_IOCTL("call ioctl:NOE_SLEEP_MODE\n");
#if 0
                int i,j;
                /*
                // ioctl_sleep
                // noe_req_sleep_on
                // noe_sleep_on
                //
                //  is very undesired implementation.
                */
	_sleep_retry:;
		for(i = 1;i < MAX_NIF;i++) {
			for(j = 0;j < MAX_NIF_SUB;j++) {
				if(noe_drv->netdev[i][j]) {
					msleep(10);
					goto _sleep_retry;
				}
			}
		}
		spin_lock_irqsave(&noe_drv->bulkout_lock, flags);
		noe_drv->noe_req_sleep_on = 1;
		noe_drv->noe_sleep_on = 1;
		oldflags = test_and_set_bit(RIGHT_TO_USE_BULKOUT, &noe_drv->flags);
		spin_unlock_irqrestore(&noe_drv->bulkout_lock, flags);
		if(oldflags)
			retval = wait_event_interruptible_timeout(noe_drv->ioctl_sleep, (noe_drv->noe_req_sleep_on == 0), SLEEP_WAIT_TIMEOUT);
		// nobody can send USB packet because RIGHT_TO_USE_BULKOUT never be release after we acquired it.
#else
                panic("NOE_DO_SLEEP_MODE is not supported");
#endif
        }
		break;

        /***************** added commands by sony *********************/

        case NOE_PRE_CLOSE_IOCTL:
            NOE_TRACE_IOCTL("call ioctl:NOE_CLOSE_IOCTL\n");
            _noe_pre_close( noe_drv );
            break;

        case NOE_INIT_IOCTL:
            NOE_TRACE_IOCTL("call ioctl:NOE_INIT_IOCTL\n");
            _noe_init( noe_drv );
            break;

        case NOE_START_SLEEP_MODE:
            NOE_INFO("NOE_START_SLEEP_MODE\n");
            set_bit( DEF_NFE_SLEEP_MODE, &noe_drv->do_sleep_flag );
            break;

        case NOE_RECEIVE_DO_SLEEP:
            NOE_TRACE_IOCTL("call ioctl:NOE_RECEIVE_DO_SLEEP\n");
	    set_bit(DEF_NFE_STATUS_DO_SLEEP_FLAG, &noe_drv->noe_status);
            wait_event_interruptible_timeout( noe_drv->do_sleep_wait_queue,
                                              test_bit( DEF_NFE_SLEEP_RECEIVE_PACKET, &noe_drv->do_sleep_flag ),
                                              DEF_NFE_WAIT_SLEEP_TIMEOUT  );
            break;

        /* DLNA server */
        case NOE_READOBJ_SEND_DATA_FROM_HIGHMEM:
        {
            int highmem_total_send_len = 0;
            struct noe_dlna_highmem_info_st *info = &noe_drv->dlna_highmem_info;

            NOE_TRACE_IOCTL("call ioctl:NOE_READOBJ_SEND_DATA_FROM_HIGHMEM\n");

	    if( down_interruptible( &noe_drv->ioctl_sem )){
	      res = -ERESTARTSYS;
	      break;
	    }
	    
            copy_from_user( info, user_buffer, sizeof(struct noe_dlna_highmem_info_st));

            NOE_DLNA_SVR_DBG("Set Shared Mem Info [add:0x%08x][size:%d][flag:%x][HostID:0x%x]\n",
                             (unsigned int)info->start_add,
                             (int)info->read_size,
                             (unsigned int)info->last_flag&0x1,
                             (unsigned int)info->HostID );
          
            _start_read_object( noe_drv, &highmem_total_send_len );
          
            info->actual_read_size = highmem_total_send_len;
            NOE_DLNA_SVR_DBG("NOE_READOBJ_SEND_DATA_FROM_HIGHMEM Done total:0x%x\n",
                             (unsigned int)info->actual_read_size );
            copy_to_user( user_buffer, info, sizeof(struct noe_dlna_highmem_info_st));

	    up( &noe_drv->ioctl_sem);
            break;
        }

        /* HDRL server */
        case NOE_HDRL_SVR_PRE_OPEN_OBJECT:
        {
          /* if we are enable to do multi hdrl, see hostid and ctx. */
            NOE_HDRL_STATUS status;

            NOE_HDRL_SVR_DBG(" [SDKTASK]<PreOpenObject>\n");
            atomic_set(&noe_drv->hdrl_write_host_id, NOE_INVALID_HOST_ID);

            status = atomic_read(&noe_drv->hdrl_status);
            if( status != NOE_HDRL_STATUS_IDLE ){
                NOE_WARNING("NOE HDRL previous status is wrong:%d", status);
            }
            atomic_set(&noe_drv->hdrl_status, NOE_HDRL_STATUS_READY);
            
            break;
        }

	    /* added for HDRL_Move */
        case NOE_HDRL_SVR_HIGHMEM_WRITE_UNIT:
	{
	    struct noe_hdrl_request_buffer_info info;
	    int ret = 0;

            copy_from_user( &info, user_buffer, sizeof(struct noe_hdrl_request_buffer_info) );

            if( atomic_read(&noe_drv->hdrl_status) != NOE_HDRL_STATUS_READY ){
                retval = 0;
                break;
            }

            ret = _enqueue_hdrl_write_unit( &noe_drv->request_write_queue, info.p_buffer, info.size );
	    if( ret != 0 ){
                retval = ret;
                break;
	    }
	      
            break;
	  }

        case NOE_HDRL_SVR_WAIT_QUEUE_READY:
        {
            retval = _wait_hdrl_queue_ready( &noe_drv->request_write_queue );
            break;
        }

        case NOE_HDRL_SVR_WAIT_UNIT_FILLED:
        {
            struct noe_hdrl_received_info info;

            retval = _wait_hdrl_unit_filled( &noe_drv->request_write_queue, &info );

            if( retval < 0 ) break;

            copy_to_user( user_buffer, &info, sizeof(struct noe_hdrl_received_info) );

            break;
        }
        /* added for HDRL_Move */

        case NOE_HDRL_SVR_SET_HOST_DATA:
        {
            NOE_UINT32 hostid;

            copy_from_user( &hostid, user_buffer, sizeof( NOE_UINT32 ) );
            atomic_set(&noe_drv->hdrl_write_host_id, hostid);
            NOE_HDRL_SVR_DBG(" [SDKTASK]<SetHostData> HostID=%x\n", hostid );


            break;
        }

        case NOE_HDRL_SVR_CLEAR_HOST_DATA:
            NOE_HDRL_SVR_DBG(" [SDKTASK]<ClearHostData>\n");
            atomic_set(&noe_drv->hdrl_write_host_id, NOE_INVALID_HOST_ID);
            break;

        /* call from receive thread by NOE_DMS_REQ_CLOSE_OBJECT_FOR_WRITE */
        case NOE_HDRL_SVR_PRE_CLOSE_OBJECT:
        {
            /* clear permit multi calling. */
            /* changed for HDRL_Move */
            NOE_HDRL_SVR_DBG(" [SDKTASK]<PreCloseObject>\n");
	    
            atomic_set(&noe_drv->hdrl_write_host_id, NOE_INVALID_HOST_ID);
            _put_write_object_cmd( &noe_drv->hdrl_write_queue, NOE_HDRL_COMMAND_FLUSH_DATA, NULL );
            atomic_set(&noe_drv->hdrl_status, NOE_HDRL_STATUS_TERMINATING);
           
            break;
        }
        case NOE_HDRL_SVR_POST_CLOSE_OBJECT:
        {
            NOE_HDRL_SVR_DBG(" [SDKTASK]<PostCloseObject>\n");
	    
	    _revoke_hdrl_request_write_queue(noe_drv);
            atomic_set(&noe_drv->hdrl_status, NOE_HDRL_STATUS_IDLE);

#if 0
            {
                request_write_queue *queue_info = &noe_drv->request_write_queue;
                printk("############################\n");
                printk("     HDRL CloseObject\n");
                printk("############################\n");
                printk("   fill_cnt         = %d\n", atomic_read(&queue_info->fill_cnt));
                printk("   fill_list empty  = %d\n", list_empty(&queue_info->fill_list));
                printk("   ready_cnt        = %d\n", atomic_read(&queue_info->ready_cnt));
                printk("   ready_list empty = %d\n", list_empty(&queue_info->ready_list));
            }
#endif
           
            break;
        }
        case NOE_HDRL_SVR_STOP_WRITE:
            /* if this flag is set, do not write high mem. */

            NOE_HDRL_SVR_DBG("<StopWrite>\n");

            break;

        case NOE_HDRL_REC_PAUSE:
            NOE_HDRL_SVR_DBG("<RecPause>\n");
            set_bit( DEF_NFE_REC_STATUS_REC_PAUSE, &noe_drv->noe_rec_status );
            break;

        case NOE_HDRL_REC_RESTART:
            NOE_HDRL_SVR_DBG("<RecRestart>\n");
            clear_bit( DEF_NFE_REC_STATUS_REC_PAUSE, &noe_drv->noe_rec_status );
            break;	

        /* others */
        case NOE_GET_PACKET_SIZE_STATUS:
            NOE_TRACE_IOCTL("call ioctl:NOE_GET_PACKET_SIZE_STATUS\n");
            copy_to_user( user_buffer, &noe_drv->noe_packet_size_status, sizeof(struct noe_packet_size_status_st) );
            break;

        case NOE_PRINT_CMD_DATA:
#ifdef NOE_CMD_DEBUG
          print_cmd_data();
#endif /*NOE_CMD_DEBUG*/
          _noe_data_dump( noe_drv, 0, 0 );
          
          break;

        case NOE_GET_FLAG_DATA:
        {
            struct noe_check_flag_data_st t_noe_check_flag_data;

            NOE_TRACE_IOCTL("call ioctl:NOE_GET_FLAG_DATA\n");

            t_noe_check_flag_data.usb_rec_flag = 0;
            t_noe_check_flag_data.net_send_flag = 0; /* sorry. not supported */
            t_noe_check_flag_data.do_sleep_flag = noe_drv->do_sleep_flag;
            t_noe_check_flag_data.hdrl_write_flag =0;
	    
            copy_to_user( user_buffer, &t_noe_check_flag_data, sizeof( struct noe_check_flag_data_st ) );

            break;
        }

        case NOE_NETWORK_MONITOR_START:
            NOE_TRACE_IOCTL("call ioctl:NOE_NETWORK_MONITOR_START\n");
            if( !test_and_set_bit( DEF_NFE_DEBUG_NETWORK_MONITOR, &noe_drv->debug_status ) ){
	        _start_network_monitor_thread(noe_drv);
            }
            break;

        case NOE_NETWORK_MONITOR_STOP:
            NOE_TRACE_IOCTL("call ioctl:NOE_NETWORK_MONITOR_STOP\n");
            if( test_and_clear_bit( DEF_NFE_DEBUG_NETWORK_MONITOR, &noe_drv->debug_status ) ){
	        _stop_network_monitor_thread(noe_drv);
            }
            break;

        case NOE_SET_CHK_STRM_FLAG:
        {
            int check_flag;

            NOE_TRACE_IOCTL("call ioctl:NOE_SET_CHK_STRM_FLAG\n");
            copy_from_user(&check_flag, user_buffer, sizeof(int));
            printk("set check stream %s\n",check_flag ? "T" : "F");

            if( check_flag ){
                set_bit(DEF_NFE_DEBUG_CHECK_STREAM, &noe_drv->debug_status);
            }
          break;
        }

        case NOE_GET_UUID_NIF:
        {
	    ioctl_get_uuid_t data;
            NOE_TRACE_IOCTL("call ioctl:NOE_GET_UUID_NIF\n");
            copy_from_user(&data, user_buffer, sizeof(ioctl_get_uuid_t));	    
	    data.ret = _get_uuid_nif( noe_drv, data.DeviceID, data.uuid, (NOE_UINT32)atomic_read(&noe_drv->current_nif) );
	    copy_to_user(user_buffer, &data, sizeof(ioctl_get_uuid_t));
	    
	    break;
	}

	case NOE_GET_NET_INFO_NIF:
        {
	    ioctl_get_net_info_t data;
            NOE_TRACE_IOCTL("call ioctl:NOE_GET_NET_INFO_NIF\n");
            copy_from_user(&data, user_buffer, sizeof(ioctl_get_net_info_t));	    
	    data.ret = _get_net_info_nif( noe_drv, data.netinfo,  (NOE_UINT32)atomic_read(&noe_drv->current_nif) );
	    copy_to_user(user_buffer, &data, sizeof(ioctl_get_net_info_t));

	    break;
	}

	case NOE_SET_NET_INFO_NIF:
        {
	    ioctl_set_net_info_t data;
            NOE_TRACE_IOCTL("call ioctl:NOE_SET_NET_INFO_NIF\n");
            copy_from_user(&data, user_buffer, sizeof(ioctl_set_net_info_t));
	    data.netinfo->nif_no = (NOE_UINT8)atomic_read(&noe_drv->current_nif);
	    data.ret = _set_net_info_nif( noe_drv, data.num_netinfo, data.netinfo );
	    copy_to_user(user_buffer, &data, sizeof(ioctl_set_net_info_t));

	    break;
	}

	case NOE_SET_CURRENT_NIF_WIRED:
	{
	    NOE_TRACE_IOCTL("call ioctl:NOE_SET_CURRENT_NIF_WIRED\n");
	    //copy_from_user(&usr_current_nif, user_buffer, sizeof(ioctl_current_nif_t));
	    if((NOE_UINT32)atomic_read(&noe_drv->current_nif) == 0){
	      break;

	    }
	    else{
	      atomic_set(&noe_drv->current_nif, 0);
	      break;
	    }
	}

	case NOE_SET_CURRENT_NIF_BSS:
	{
	    NOE_TRACE_IOCTL("call ioctl:NOE_SET_CURRENT_NIF_BSS\n");
	    //copy_from_user(&usr_current_nif, user_buffer, sizeof(ioctl_current_nif_t));
	    if((NOE_UINT32)atomic_read(&noe_drv->current_nif) == 1){
	      break;
	    }
	    else{
	      atomic_set(&noe_drv->current_nif, 1);
	      break;
	    }
	}


	default:
		retval = -EOPNOTSUPP;
		break;
	}

	return(retval);
}

static struct file_operations noe_fops = {
	.owner =	THIS_MODULE,
	.read =		noe_read,
	.write =	noe_write,
	.ioctl =	noe_ioctl,
	.open =		noe_open,
	.release =	noe_release,
};

/*
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with the driver core
 */
static struct usb_class_driver noe_class = {
	.name =		"noe%d",
	.fops =		&noe_fops,
	.minor_base =	NOE_USB_MINOR_BASE,
};

static int
noe_suspend(struct usb_interface *interface, pm_message_t message)
{
	int	i, j;
	int	running = 0;
	noe_drv_t *dev = usb_get_intfdata(interface);

	NOE_INFO("ether_suspend\n");

	for(i = 0;i < MAX_NIF;i++) {
		for(j = 0;j < MAX_NIF_SUB;j++) {
                        struct net_device *netdev = netdevinfo2netdev( dev, i, j );
			if(netdev) {
				netif_device_detach(netdev);
				if(netif_running(netdev))
					running = 1;
			}
		}
	}

	if(running)
		usb_kill_urb(dev->read_urb);

	return(0);
}

static int
noe_resume(struct usb_interface *interface)
{
	int	i, j;
	int	running = 0;
	noe_drv_t *dev = usb_get_intfdata(interface);

	NOE_INFO("ether_resume\n");

	for(i = 0;i < MAX_NIF;i++) {
		for(j = 0;j < MAX_NIF_SUB;j++) {
                        struct net_device *netdev = netdevinfo2netdev( dev, i, j );
			if(netdev) {
				netif_device_attach(netdev);
				if(netif_running(netdev))
					running = 1;
			}
		}
	}

	if(running) {
		dev->read_urb->status = 0;
		dev->read_urb->actual_length = 0;
		ether_read_callback(dev->read_urb);
	}

	return(0);
}


static void
noe_delete(struct kref *kref)
{
	noe_drv_t *dev = container_of(kref, noe_drv_t, kref);
        noe_packet_buffer_t *c, *n;

	NOE_DP("noe_delete\n");

	usb_kill_urb(dev->read_urb);
	usb_kill_urb(dev->write_urb);
	usb_kill_urb(dev->ctrl_urb);

	usb_put_dev(dev->udev);

        /*********** modified by sony ************/
        list_for_each_entry_safe( c, n, &dev->packet_buffer_by_hostid.free_list, list ){
                list_del(&c->list);
                kfree(c);
        }
        list_for_each_entry_safe( c, n, &dev->packet_buffer_by_hostid.used_list, list ){
                if(c->data){
                    net_drv_free_rx_buf(dev, c->data);
                }
                c->data = NULL;
                list_del(&c->list);
                kfree(c);
	}
        /*****************************************/
	if(dev->bulk_out_buffer)
		kfree(dev->bulk_out_buffer);
	if(dev->read_obj_out_buffer)
		kfree(dev->read_obj_out_buffer);
	if(dev->write_urb)
		usb_free_urb(dev->write_urb);
	if(dev->read_urb)
		usb_free_urb(dev->read_urb);
	if(dev->ctrl_urb)
		usb_free_urb(dev->ctrl_urb);

	kfree(dev);
}

int
noe_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	noe_drv_t	*dev = NULL;
	struct usb_host_interface	*iface_desc;
	struct usb_endpoint_descriptor	*endpoint;
        struct noe_packet_size_status_st *pksize_status; /* added by sony */
        hdrl_write_queue                 *hdrl_write_queue; /* added by sony */
	request_write_queue              *request_write_queue; /* added for HDRL_Move */
	int	i,j;
	int	retval = -ENOMEM;

	NOE_DP("start_noe_probe \n");

	/* allocate memory for our device state and initialize it */
	dev = kzalloc(sizeof(*dev), GFP_KERNEL);
	if(dev == NULL) {
		NOE_ERROR("Out of memory\n");
		return(-ENOMEM);
	}
	kref_init(&dev->kref);

	mutex_init(&dev->write_mutex);

	init_waitqueue_head(&dev->ioctl_noe);

	for(i = 0;i < WQ_RINGS;i++) {
		init_waitqueue_head(&dev->wait_queue[i].ioctl_wait);
		dev->wait_queue[i].HostID = NOE_HID_NONE;
	}
	spin_lock_init(&dev->rd_ioctl_lock);

	init_waitqueue_head(&dev->bulkout_wait);

	atomic_set(&dev->ether_opened, 0);

	dev->udev = usb_get_dev(interface_to_usbdev(interface));

        hdrl_write_queue = &dev->hdrl_write_queue; /* added by sony */
        request_write_queue = &dev->request_write_queue; /* added for HDRL_Move */	

	/* set up the endpoint information */
	/* use only the first bulk-in and bulk-out endpoints */
	iface_desc = interface->cur_altsetting;
	for(i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if(!dev->bulk_in_endpointAddr && usb_endpoint_is_bulk_in(endpoint)) {
                        noe_packet_buffer_list_t *buffer_list = &dev->packet_buffer_by_hostid;
			/* we found a bulk in endpoint */
			dev->read_urb = usb_alloc_urb(0, GFP_KERNEL);
			if(!dev->read_urb) {
				NOE_ERROR("No free read_urbs available\n");
				goto _error;
			}
			dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;

                        /* changed buffer structure by sony */
                        INIT_LIST_HEAD(&buffer_list->used_list);
#ifdef DEBUG_PACKET_BUFFER_LIST
                        buffer_list->used_cnt = 0;
#endif
                        INIT_LIST_HEAD(&buffer_list->free_list);
			for(j = 0;j < NOE_RINGS;j++) {
                                noe_packet_buffer_t *packet_buffer;
                                packet_buffer = kmalloc(sizeof(noe_packet_buffer_t), GFP_KERNEL);
                                if(!packet_buffer){
                                    NOE_ERROR("cannot allocate packet_buffer\n");
                                    goto _error;
                                }

				packet_buffer->data = NULL;
				packet_buffer->HostID = NOE_HID_NONE;

                                list_add_tail(&(packet_buffer->list), &(buffer_list->free_list));
#ifdef DEBUG_PACKET_BUFFER_LIST
                                buffer_list->free_cnt++;
#endif
			}
		}

		if(!dev->bulk_out_endpointAddr && usb_endpoint_is_bulk_out(endpoint)) {
			/* we found a bulk out endpoint */
			dev->write_urb = usb_alloc_urb(0, GFP_KERNEL);
			if(!dev->write_urb) {
				NOE_ERROR("No free write_urbs available\n");
				goto _error;
			}
			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
			dev->bulk_out_buffer = kmalloc(NOE_BUFSIZE, GFP_KERNEL);
			if(!dev->bulk_out_buffer) {
				NOE_ERROR("Couldn't allocate bulk_out_buffer\n");
				goto _error;
			}
                        /* added by sony */
			dev->read_obj_out_buffer = kmalloc(NOE_BUFSIZE, GFP_KERNEL);
			if(!dev->read_obj_out_buffer) {
				NOE_ERROR("Couldn't allocate read_obj_out_buffer\n");
				goto _error;
			}
		}
	}
	dev->ctrl_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!dev->ctrl_urb) {
		NOE_ERROR("No free ctrl_urbs available\n");
		goto _error;
	}

	if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) {
		NOE_ERROR("Could not find both bulk-in and bulk-out endpoints\n");
		goto _error;
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);

	/* we can register the device now, as it is ready */
	retval = usb_register_dev(interface, &noe_class);
	if (retval) {
		/* something prevented us from registering this driver */
		NOE_ERROR("Not able to get a minor for this device.\n");
		usb_set_intfdata(interface, NULL);
		goto _error;
	}

//	printk(DRIVER_DESC " " DRIVER_VERSION "\n");
	/* let the user know what node this device is now attached to */
	NOE_INFO("usb_drv device now attached to #%d\n", interface->minor - NOE_USB_MINOR_BASE);

	dev->interface = interface;

        /****** added by sony *****/
        pksize_status = &dev->noe_packet_size_status;

        if( g_noe_usb_mtu_num > NFE_MAX_MULTI_PACKET_CNT || g_noe_usb_mtu_num < 1 ){
            pksize_status->mtu_num = NFE_DEFAULT_MULTI_PACKET_CNT;
        }
        else {
            pksize_status->mtu_num = g_noe_usb_mtu_num;
        }

        pksize_status->recv_payload_size = HOST_DRIVER_MAXPACKET*dev->noe_packet_size_status.mtu_num;
        pksize_status->recv_packet_size = dev->noe_packet_size_status.recv_payload_size  + sizeof(NOE_HEADER);
        
        pksize_status->send_payload_size = HOST_DRIVER_MAXPACKET - sizeof(NOE_HEADER);
        pksize_status->send_packet_size = HOST_DRIVER_MAXPACKET;
        
        printk("[NFEUSB]usb_drv mtu:%d  <recv>payload:%d packet:%d <send>payload:%d packet:%d\n",
               (int)pksize_status->mtu_num,
               (int)pksize_status->recv_payload_size,
               (int)pksize_status->recv_packet_size,
               (int)pksize_status->send_payload_size,
               (int)pksize_status->send_packet_size );

        dev->noe_status = 0;
        dev->noe_thread_status = 0;
        dev->noe_rec_status = 0;
        dev->debug_status = 0;
        dev->debug_pass_thru_rcv_len = 0;
        dev->debug_pass_thru_snd_len = 0;
        dev->debug_dlna_rcv_len = 0;
        dev->debug_noe_data_snd_len = 0;
        dev->debug_hdrl_write_notify_cnt = 0;

	atomic_set(&dev->current_nif,0);/*add by sony*/

	dev->current_console_para = NOE_MONITOR_DISABLE;

        INIT_CMD_HISTORY(dev);

        set_bit( DEF_NFE_USB_STATUS_INIT, &g_noe_usb_status );

        _init_netdev_info(dev);

        init_MUTEX(&dev->noe_rcv_thd_data_sem);
	init_MUTEX(&dev->ioctl_sem);
	
        INIT_LIST_HEAD(&dev->noe_rcv_thd_data_list);
        spin_lock_init(&dev->noe_rcv_thd_data_list_lock);
        atomic_set(&dev->noe_rcv_thd_data_list_cnt, 0);
        init_waitqueue_head(&dev->noe_rcv_thd_data_wait_queue);
        dev->noe_rcv_thd_buffer = kmem_cache_create("noe_rcv_thd_cache", sizeof(noe_rcv_thd_buffer_t),
                                                    0, SLAB_HWCACHE_ALIGN, NULL );

        init_waitqueue_head(&dev->do_sleep_wait_queue);
        dev->do_sleep_flag = 0;
        dev->do_sleep_hostid = NOE_INVALID_HOST_ID;

	/********************************/
        /*     HDRL                     */
	/********************************/
        atomic_set(&dev->hdrl_status, NOE_HDRL_STATUS_IDLE);

        hdrl_write_queue->cache = kmem_cache_create("noe_hdrl_write_cache", sizeof(noe_hdrl_data_t),
                                                    0, SLAB_HWCACHE_ALIGN, NULL );
        atomic_set(&dev->hdrl_write_host_id, NOE_INVALID_HOST_ID);
        spin_lock_init(&hdrl_write_queue->list_lock);
        INIT_LIST_HEAD(&hdrl_write_queue->list);
        atomic_set(&hdrl_write_queue->cnt, 0);
        init_waitqueue_head(&hdrl_write_queue->wq);

	request_write_queue->cache = kmem_cache_create("noe_hdrl_request_write_cache", sizeof(noe_write_unit_data_t),
						       0, SLAB_HWCACHE_ALIGN, NULL );
        spin_lock_init(&request_write_queue->list_lock);
        INIT_LIST_HEAD(&request_write_queue->ready_list);
        atomic_set(&request_write_queue->ready_cnt, 0);
        init_waitqueue_head(&request_write_queue->ready_wq);
        INIT_LIST_HEAD(&request_write_queue->fill_list);
        atomic_set(&request_write_queue->fill_cnt, 0);
        init_waitqueue_head(&request_write_queue->fill_wq);

        dev->write_object_thread = _create_write_object_thread(dev);





        dev->noe_send_status = 0;
        dev->vap_status = 0;

        _setup_procfs(dev);
	_setup_procfs_nif(dev);

	retval = ether_probe(dev);
	if(retval < 0)
		goto _error;

	return(0);

_error:;

        clear_bit( DEF_NFE_USB_STATUS_INIT, &g_noe_usb_status );

        if(dev->noe_rcv_thd_buffer){
            kmem_cache_destroy(dev->noe_rcv_thd_buffer);
        }

        if(dev->write_object_thread){
            _destroy_write_object_thread(dev);
        }

        if(hdrl_write_queue->cache){
            kmem_cache_destroy(hdrl_write_queue->cache);
        }
	/* added for HDRL_Move */
        if(request_write_queue->cache){
            kmem_cache_destroy(request_write_queue->cache);
        }
	/* */
	
	if(dev)
		kref_put(&dev->kref, noe_delete);

	_terminate_procfs_nif();
	
	return(retval);
}

void noe_disconnect(struct usb_interface *interface)
{
	noe_drv_t *dev;

	NOE_INFO("usb_drv #%d now disconnected\n", interface->minor - NOE_USB_MINOR_BASE);

	dev = usb_get_intfdata(interface);

        clear_bit( DEF_NFE_USB_STATUS_INIT, &g_noe_usb_status ); /* added by sony */

	ether_disconnect(interface);

	usb_set_intfdata(interface, NULL);

	/* give back our minor */
	usb_deregister_dev(interface, &noe_class);

	/* decrement our usage count */
	kref_put(&dev->kref, noe_delete);

#if defined(__UBI_DEBUG) && __UBI_DEBUG
	noe_host_seq = 0;
	noe_subsys_sequence = 0;
#endif

        /****** added by sony *****/
        _terminate_procfs(dev);

        if(dev->noe_rcv_thd_buffer){
            NOE_DP(" destory rcv_thd_buffer.......\n");
            kmem_cache_destroy(dev->noe_rcv_thd_buffer);
        }

        NOE_DP(" destory write_object thread.......\n");
        _destroy_write_object_thread(dev);
        NOE_DP(" destory write_object thread.......done\n");

        if(dev->hdrl_write_queue.cache){
            NOE_DP(" destory hdrl_write_cache.......\n");
            kmem_cache_destroy(dev->hdrl_write_queue.cache);
        }

	/* added for HDRL_Move */
        if(dev->request_write_queue.cache){
            NOE_DP(" destory request_write_cache.......\n");
            kmem_cache_destroy(dev->request_write_queue.cache);
        }

	_terminate_procfs_nif();
}

/*****************************************************************/
/*                   added functions by sony                     */
/*****************************************************************/
static inline bool _is_noe_rcv_thd( NOE_UINT32 HostID )
{
    return ( NOE_HID_UNBIND == HostID );
}

/***********************************/
/*     NOE Thread Data control     */
/***********************************/
static int _send_noe_rcv_thd_data( noe_drv_t *noe_drv, noe_rcv_usb_data *data )
{
    noe_rcv_thd_buffer_t *c;
    unsigned long         flags;

    c = (noe_rcv_thd_buffer_t*) kmem_cache_alloc( noe_drv->noe_rcv_thd_buffer, GFP_ATOMIC );
    if( c == NULL ){
        NOE_ERROR("cannot allocate noe_rcv_thd_data\n");
        return -ENOMEM;
    }

    c->data = data;

    spin_lock_irqsave(&noe_drv->noe_rcv_thd_data_list_lock, flags);
    list_add_tail(&c->list, &noe_drv->noe_rcv_thd_data_list);
    atomic_inc(&noe_drv->noe_rcv_thd_data_list_cnt);
    spin_unlock_irqrestore(&noe_drv->noe_rcv_thd_data_list_lock, flags);

    if( waitqueue_active(&noe_drv->noe_rcv_thd_data_wait_queue) ){
        wake_up_interruptible(&noe_drv->noe_rcv_thd_data_wait_queue);
    }

    return 0;
}

static int _receive_noe_rcv_thd_data( noe_drv_t *noe_drv, noe_rcv_usb_data **data )
{
    noe_rcv_thd_buffer_t *c;
    unsigned long         flags;
    int                   retval;

    retval = wait_event_interruptible( noe_drv->noe_rcv_thd_data_wait_queue,
                                       atomic_read(&noe_drv->noe_rcv_thd_data_list_cnt) > 0 );
    if( retval < 0 )return retval;

    spin_lock_irqsave(&noe_drv->noe_rcv_thd_data_list_lock, flags);
    c = list_first_entry( &noe_drv->noe_rcv_thd_data_list, noe_rcv_thd_buffer_t, list );
    *data = c->data;
    list_del(&c->list);
    kmem_cache_free( noe_drv->noe_rcv_thd_buffer, c );
    atomic_dec(&noe_drv->noe_rcv_thd_data_list_cnt);
    spin_unlock_irqrestore(&noe_drv->noe_rcv_thd_data_list_lock, flags);

    return 0;
}

/************************/
/*       for ioctl      */
/************************/
/* control through ioctl:NOE_READ_IOCTL */
static int _get_noe_rcv_thd_data( noe_drv_t *noe_drv, void *user_buffer, NOE_pFRAME *noe_packet )
{
    noe_rcv_usb_data*   data = NULL;
    int                 retval = 0;
    int                 copy_st;
    int                 len;

    /* rcv_thd permits only one thread */
    if( down_interruptible(&noe_drv->noe_rcv_thd_data_sem) ){
        return -ERESTARTSYS;
    }

    while(1){
        if( test_bit(DEF_NFE_STATUS_PRE_CLOSE, &noe_drv->noe_status) ){
            /* at PreClose */
            NOE_HEADER close_h;
          
            memset(&close_h, 0x0, sizeof(NOE_HEADER));
            close_h.cmd = NOE_USER_RCV_THD_CLOSE_COMMAND;
            copy_to_user(user_buffer, &close_h, sizeof(NOE_HEADER));
            clear_bit(DEF_NFE_STATUS_PRE_CLOSE, &noe_drv->noe_status);
            break;
        }
        else{
            /* at normal */
            NOE_pFRAME* frame;

            retval = _receive_noe_rcv_thd_data(noe_drv, &data);
            if( retval < 0 ) break;

            if( data == NULL ) continue; /* dummy wakeup */

            frame = GET_RCV_USB_DATA_BODY(data);

            len = frame->h.len;
            copy_st = copy_to_user(user_buffer, &(frame->h), sizeof(NOE_HEADER));
            if(copy_st == 0 && len)
              copy_st = copy_to_user(noe_packet->payload, (void*)(&(frame->h)) + sizeof(NOE_HEADER), len);

            net_drv_free_rx_buf(noe_drv, data);
            break;
        }
    }

    up(&noe_drv->noe_rcv_thd_data_sem);

    return retval;
}

static int _read_noe_data( noe_drv_t *noe_drv, NOE_UINT32 HostID, void* user_header, void *user_payload )
{
    noe_wait_queue_t	*wq;
    noe_packet_buffer_t	*pb;
    unsigned long	copy_st;
    unsigned long       flags;
    int                 retval = 0;
    int                 len;
    NOE_pFRAME          *frame;

    while(1) {
        spin_lock_irqsave(&noe_drv->rd_ioctl_lock, flags);
        pb = _find_packet(&noe_drv->packet_buffer_by_hostid, HostID);
        if(pb){
            spin_unlock_irqrestore(&noe_drv->rd_ioctl_lock, flags);
            break;
        }

        wq = get_free_wait_queue(noe_drv->wait_queue);
        if(wq == NULL) {
            spin_unlock_irqrestore(&noe_drv->rd_ioctl_lock, flags);
            NOE_ERROR("no more wait queue\n");
            return(-EBUSY);
        }
        wq->HostID = HostID;
        spin_unlock_irqrestore(&noe_drv->rd_ioctl_lock, flags);

        retval = wait_event_interruptible(wq->ioctl_wait, (wq->HostID == NOE_HID_NONE));
        if(retval)
            return(retval);
    }

    frame = GET_RCV_USB_DATA_BODY(pb->data);

    len = sub2host32(GET_RCV_USB_DATA_BODY(pb->data)->h.len);
    copy_st = copy_to_user( user_header, &(frame->h), sizeof(NOE_HEADER));
    if(copy_st == 0 && len){
        void *payload = (void*)(&(frame->h)) + sizeof(NOE_HEADER);
        copy_st = copy_to_user(user_payload, payload, len);
    }

    spin_lock_irqsave(&noe_drv->rd_ioctl_lock, flags);
    _release_used_packet( noe_drv, pb );
    spin_unlock_irqrestore(&noe_drv->rd_ioctl_lock, flags);

    return 0;
}

static void _noe_pre_close( noe_drv_t *noe_drv )
{
    set_bit(DEF_NFE_STATUS_PRE_CLOSE, &noe_drv->noe_status);
    _send_noe_rcv_thd_data( noe_drv, NULL ); /* dummy wakeup */
}

static void _noe_init( noe_drv_t *noe_drv )
{
  /*    clear_bit( DEF_NFE_NET_SEND_SLEEP_ON, &dev->net_send_flag ); */
     clear_bit(DEF_NFE_STATUS_DO_SLEEP_FLAG, &noe_drv->noe_status);

    noe_drv->do_sleep_flag = 0;
    noe_drv->do_sleep_hostid = NOE_INVALID_HOST_ID;
    noe_drv->debug_status = 0;
}

/************************/
/*   HighMem control    */
/************************/
/* added for HDRL_Move */
static int _copy_noe_data_to_highmem( noe_drv_t *noe_drv, NOE_pFRAME *frame )
{
    void *t_buf = NULL;
    NOE_HEADER *noe_header = NULL;
    unsigned long write_len = 0;
    int ret = 0;
    bool ack = false;

    if( frame == NULL ){
        /*nothting to do.*/
        return 0;
    }

    noe_header= &frame->h;
    write_len = noe_header->len;
    
    if( noe_header->cmd != NOE_ACK_MORE ){
        NOE_HDRL_SVR_DBG("Receive ACK!!\n");
        ack = true;
    }

    /* changed for HDRL_Move */
    t_buf = (void*)frame;
    t_buf += sizeof(NOE_HEADER);

    _fill_hdrl_write_unit( &noe_drv->request_write_queue, t_buf, write_len, ack );

    return ret;
}

/* added for HDRL_Move */
static unsigned long _get_highmem_copy_length( noe_drv_t *noe_drv,
					       unsigned long highmem_total_send_len )
{
    unsigned long len = noe_drv->dlna_highmem_info.read_size - highmem_total_send_len;
    unsigned long align_len = NOE_MAKEALIGN(len);

    if( align_len > noe_drv->noe_packet_size_status.send_payload_size ){
	align_len = noe_drv->noe_packet_size_status.send_payload_size;
    }

    NOE_DLNA_SVR_DBG("_get_highmem_copy_length req_size=%d total_send=%d now_len=%d align_len=%d\n",
		     (int)noe_drv->dlna_highmem_info.read_size,
		     (int)highmem_total_send_len,(int)len, (int)align_len);
    return align_len;
}

static unsigned long _copy_data_from_highmem( noe_drv_t *noe_drv,
                                              void *buffer, unsigned long highmem_total_send_len )
{
    unsigned long length;
    unsigned long start_address = noe_drv->dlna_highmem_info.start_add + highmem_total_send_len;
    NOE_DLNA_SVR_DBG("[_copy_data_from_highmem] add=%x total_send_size=%d \n",
		     (unsigned int)noe_drv->dlna_highmem_info.start_add,
		     (int)highmem_total_send_len);

    length = _get_highmem_copy_length( noe_drv, highmem_total_send_len );

    NOE_DLNA_SVR_DBG("[_copy_data_from_highmem] memcpy src:%x dst:%x len=%d\n",
		     (unsigned int)start_address,
		     (unsigned int)((void*)noe_drv->bulk_out_buffer + sizeof( NOE_HEADER)),
		     (int)length );

    memcpy( buffer + sizeof( NOE_HEADER ), (void*)start_address, length );
    return length;
}

/************************/
/*     DLNA control     */
/************************/
static void _make_read_object_header( struct noe_dlna_highmem_info_st *info,
                                      void *buffer, unsigned long len, int finish )
{
    NOE_HEADER h;

    memset( &h, 0x0, sizeof(NOE_HEADER) );
   
    h.HostID = info->HostID;
    h.SubSysID = info->SubSysID;
    h.len = (NOE_UINT32)len;
    
    /*this request is last && this send is last.*/
    if( finish  ){
	h.cmd = NOE_ACK;
    } else {
	h.cmd = NOE_ACK_MORE;
    }
    h.r0 = info->read_size;

    NOE_DLNA_SVR_DBG("[_make_read_object_header]header[host:0x%x sub:0x%x len:%d cmd:%s r0=%d\n",
		     (unsigned int)h.HostID,
		     (unsigned int)h.SubSysID, (int)h.len,
		     h.cmd == NOE_ACK ? "NOE_ACK" : "NOE_ACK_MORE" ,
		     (unsigned int)h.r0 );
    
    memcpy( buffer, &h, sizeof( NOE_HEADER ) );
}

static int _start_read_object( noe_drv_t *noe_drv, int *highmem_total_send_len )
{
    int ret = 0;
    unsigned long length;  /*data length one time.*/
    unsigned long noe_align_length;
    int request_send_device_finish = 0; /* send all data for device, 1*/
    struct noe_dlna_highmem_info_st *info = &noe_drv->dlna_highmem_info;
    void *buffer = noe_drv->read_obj_out_buffer;
    struct net_device *netdev;

    mutex_lock(&noe_drv->write_mutex);

    *highmem_total_send_len = 0;

    netdev = get_current_netdev(noe_drv);
    netif_stop_queue(netdev); 

    if( wait_event_interruptible( noe_drv->bulkout_wait, !test_and_set_bit(RIGHT_TO_USE_BULKOUT, &noe_drv->flags) ) < 0 ){
        netif_wake_queue(netdev);
        return -ERESTARTSYS;
    }

    while( *highmem_total_send_len < info->read_size ){
        length = _copy_data_from_highmem( noe_drv, buffer, *highmem_total_send_len );

	request_send_device_finish = 0;
	if( *highmem_total_send_len + length >= info->read_size ){
	    request_send_device_finish = 1;
	}

        _make_read_object_header( info, buffer, length, request_send_device_finish );

	noe_align_length = NOE_MAKEALIGN(length) + sizeof(NOE_HEADER);

        ret = _noe_send_data( noe_drv, buffer, noe_align_length );
	if( ret < 0 ){
	    NOE_ERROR("_send_data_for_noe error \n");
	    break;
	}
	*highmem_total_send_len += length;

    }

    clear_bit(RIGHT_TO_USE_BULKOUT, &noe_drv->flags);
    if(waitqueue_active( &noe_drv->bulkout_wait )){
        wake_up_interruptible( &noe_drv->bulkout_wait );
    }

    netif_wake_queue(netdev);


    mutex_unlock(&noe_drv->write_mutex);

    return ret;
}

/************************/
/*     HDRL control     */
/************************/
static bool _is_write_object_cmd( noe_drv_t *noe_drv, NOE_HEADER* buf )
{
    NOE_UINT32 cmd;
    NOE_UINT32 HostID;

    cmd    = sub2host32(buf->cmd);
    HostID = sub2host32(buf->HostID);

    /* This is Write Object */
    if( ( cmd == NOE_ACK || cmd == NOE_ACK_MORE ) &&
        ( HostID == atomic_read(&noe_drv->hdrl_write_host_id) ) ) return true;

    return false;
}

static int _put_write_object_cmd( hdrl_write_queue *queue, NOE_HDRL_COMMAND command, noe_rcv_usb_data* data )
{
    unsigned long flags;
    noe_hdrl_data_t *hdrl_data;

    NOE_HDRL_SVR_DBG("Write Object Received.....\n");

    hdrl_data = (noe_hdrl_data_t*) kmem_cache_alloc( queue->cache, GFP_ATOMIC );
    if( hdrl_data == NULL ){
        return -ENOMEM;
    }
    hdrl_data->command = command;
    hdrl_data->data = data;

    spin_lock_irqsave(&queue->list_lock, flags);
        
    list_add_tail( &hdrl_data->list, &queue->list );
    atomic_inc( &queue->cnt );
    if( waitqueue_active(&queue->wq) ){
        wake_up_interruptible(&queue->wq);
    }

    spin_unlock_irqrestore(&queue->list_lock, flags);

    return 0;
}

static int _get_write_object_cmd( hdrl_write_queue *queue, NOE_HDRL_COMMAND *command, noe_rcv_usb_data** data )
{
    unsigned long flags;
    noe_hdrl_data_t *c;

    spin_lock_irqsave(&queue->list_lock, flags);

    if( list_empty(&queue->list) ){
        spin_unlock_irqrestore(&queue->list_lock, flags);
        return -ENOENT;
    }

    c = list_first_entry( &queue->list, noe_hdrl_data_t, list );
    *data = c->data;
    *command = c->command;
    list_del(&c->list);
    kmem_cache_free( queue->cache, c );
    atomic_dec( &queue->cnt );

    spin_unlock_irqrestore(&queue->list_lock, flags);

    return 0;
}

/*****************************************************
    HDRL WriteObject Procedure
*****************************************************/
static int _write_object_thread_main(void* data)
{
    noe_drv_t *noe_drv = (noe_drv_t*)data;
    hdrl_write_queue *queue;

    complete( &noe_drv->write_object_thread_start_complete );

    queue = &noe_drv->hdrl_write_queue;

    while(1){
        noe_rcv_usb_data *data = NULL;
        NOE_HDRL_COMMAND  command;
        int               write_len;
        bool              terminate = false;

        /* wait until _put_write_object_cmd */
        wait_event_interruptible( queue->wq, (atomic_read(&queue->cnt) > 0) );
        NOE_HDRL_SVR_DBG(" [KTHREAD]Start Write Object<%d> Sequence.....\n", atomic_read(&queue->cnt));
        _get_write_object_cmd( queue, &command, &data );

        switch(command){
            case NOE_HDRL_COMMAND_WRITE_DATA:
            {
                /* copy data from buffer to high-mem */
                write_len = _copy_noe_data_to_highmem( noe_drv, GET_RCV_USB_DATA_BODY(data) );

#ifdef NOE_DEF_DEBUG_CHECK_STREAM
                if( test_bit(DEF_NFE_DEBUG_CHECK_STREAM, &noe_drv->debug_status) ){
                    _hdrl_check_stream( noe_drv, buf );
                }
#endif
                /* free buffer */
                net_drv_free_rx_buf( noe_drv, data );
                break;
            }
            case NOE_HDRL_COMMAND_FLUSH_DATA:
            {
                _flush_hdrl_write_unit(&noe_drv->request_write_queue);
                break;
            }
            case NOE_HDRL_COMMAND_TERMINATE:
            {
                terminate = true;
                break;
            }
            default:
                NOE_ERROR("unknown HDRL COMMAND:%d\n", command);
                continue;
        }

        if(terminate) break;

    }

    complete( &noe_drv->write_object_thread_finish_complete );

    return 0;
}

static struct task_struct* _create_write_object_thread( noe_drv_t *noe_drv )
{
    struct task_struct *thread;
    struct sched_param sched_param;

    init_completion(&noe_drv->write_object_thread_start_complete);
    init_completion(&noe_drv->write_object_thread_finish_complete);

    thread = kthread_run( _write_object_thread_main, noe_drv, "noe_write_object" );
    if( thread == NULL ){
        NOE_ERROR("couldn't create write_object_thread\n");
        return NULL;
    }

    wait_for_completion( &noe_drv->write_object_thread_start_complete );

    sched_param.sched_priority = NOE_WRITE_OBJECT_THREAD_PRIORITY;
    if( sched_setscheduler( thread, NOE_WRITE_OBJECT_THREAD_POLICY, &sched_param ) < 0 ){
        NOE_ERROR("couldn't change priority for write_object_thread\n");
        _destroy_write_object_thread( noe_drv );
        return NULL;
    }

    return thread;
}

static void _destroy_write_object_thread( noe_drv_t *noe_drv )
{
    _put_write_object_cmd( &noe_drv->hdrl_write_queue, NOE_HDRL_COMMAND_TERMINATE, NULL );
    wait_for_completion( &noe_drv->write_object_thread_finish_complete );
}

/**************************************/
/*   HighMem control for HDRL_Move    */
/**************************************/
static int _enqueue_hdrl_write_unit( request_write_queue *queue_info, void* buffer, int size )
{
    unsigned long flags;
    noe_write_unit_data_t *unit;
    struct noe_hdrl_received_info  *info;
    struct _hdrl_received_data     *data;

    unit = (noe_write_unit_data_t*) kmem_cache_alloc( queue_info->cache, GFP_KERNEL );
    if( unit == NULL ){
        NOE_ERROR(" cannot enqueue memory unit for HDRL\n");
        return -ENOMEM;
    }

    info = &(unit->received_info);
    info->command = HDRL_REPLY_CODE_INVALID;
    data = &(info->complete_data);
    data->p_buffer = buffer;
    data->size = size;
    data->actual_size = 0;

    spin_lock_irqsave(&queue_info->list_lock, flags);
    list_add_tail( &unit->list, &queue_info->ready_list );
    atomic_inc( &queue_info->ready_cnt );
    spin_unlock_irqrestore(&queue_info->list_lock, flags);

    NOE_HDRL_SVR_DBG(" [SDKTASK] <%03d>ENQ [0x%p][%d]\n", atomic_read(&queue_info->ready_cnt), buffer, size );

    if(waitqueue_active(&queue_info->ready_wq)){
       wake_up(&queue_info->ready_wq);
    }

    return 0;
}

static int _fill_hdrl_write_unit( request_write_queue *queue_info, void* data, int size, bool ack )
{
    unsigned long flags;
    noe_write_unit_data_t *c, *n;
    int write_size;
    int total_remain_size = size;
    void *src = data;

    /* wait queue is registered */
    while(1){
        bool fill_complete = false;

        if( wait_event_interruptible( queue_info->ready_wq, (atomic_read(&queue_info->ready_cnt)>0) ) < 0 ){
            NOE_ERROR("dequeue failed for signal receiving\n");
            continue;
        }

        spin_lock_irqsave(&queue_info->list_lock, flags);
        list_for_each_entry_safe( c, n, &queue_info->ready_list, list ){

            struct noe_hdrl_received_info *received_info = &(c->received_info);
            struct _hdrl_received_data    *received_data = &(received_info->complete_data);

            int unit_remain_size = received_data->size - received_data->actual_size;
            void* dst = (void*)((char*)received_data->p_buffer + received_data->actual_size);

            /* fill data */
            write_size = (total_remain_size < unit_remain_size) ? total_remain_size : unit_remain_size;
            NOE_HDRL_SVR_DBG(" [KTHREAD] <%03d>DEQ (%d/%d) -> [0x%p/0x%p][%d/%d]\n", atomic_read(&queue_info->ready_cnt),
                             write_size, size,   dst, received_data->p_buffer,   (received_data->size-unit_remain_size), received_data->size );
            memcpy( dst, src, write_size );
            total_remain_size -= write_size;
            received_data->actual_size += write_size;
            src += write_size;
            
            if( received_data->size <= received_data->actual_size ){
                /* fill complete */
                /* dequeue from ready list and enqueue to fill list */
                received_info->command = HDRL_REPLY_CODE_DATA_COMPLETE;
                list_move_tail(&c->list, &queue_info->fill_list );
                atomic_dec( &queue_info->ready_cnt );
                atomic_inc(&queue_info->fill_cnt);
            }

            if( total_remain_size <= 0 ){
                fill_complete = true;
                break;
            }
        }
        spin_unlock_irqrestore(&queue_info->list_lock, flags);

        if(fill_complete){
            break;
        }
    }

    if(ack){
        /* ACK is received. so notify information only for ACK */
        noe_write_unit_data_t *ack_unit;

        ack_unit = (noe_write_unit_data_t*) kmem_cache_alloc( queue_info->cache, GFP_KERNEL );
        if( ack_unit == NULL ){
            NOE_ERROR("No Memory! cannot notify ACK information\n");
        }
        else{
            ack_unit->received_info.command = HDRL_REPLY_CODE_ACK_RECEIVED;
            ack_unit->received_info.ack_info = 0; /*dummy*/
        }
        spin_lock_irqsave(&queue_info->list_lock, flags);
        list_add_tail( &ack_unit->list, &queue_info->fill_list );
        atomic_inc(&queue_info->fill_cnt);
        spin_unlock_irqrestore(&queue_info->list_lock, flags);
    }

    if( (atomic_read(&queue_info->fill_cnt) > 0) && waitqueue_active(&queue_info->fill_wq) ){
        wake_up(&queue_info->fill_wq);
    }

    return 0;
}

static int _flush_hdrl_write_unit( request_write_queue *queue_info )
{
    unsigned long flags;
    noe_write_unit_data_t *c;
    struct noe_hdrl_received_info *received_info;
    struct _hdrl_received_data    *received_data;

    NOE_HDRL_SVR_DBG(" [KTHREAD] FLUSH HDRL WRITE UNIT\n");

    /* wait queue is registered */
    if( wait_event_interruptible( queue_info->ready_wq, (atomic_read(&queue_info->ready_cnt)>0) ) < 0 ){
        NOE_ERROR("dequeue failed for signal receiving\n");
        return -1;
    }

    spin_lock_irqsave(&queue_info->list_lock, flags);
    c = list_first_entry( &queue_info->ready_list, noe_write_unit_data_t, list );

    received_info = &(c->received_info);
    received_data = &(received_info->complete_data);

    received_info->command = HDRL_REPLY_CODE_DATA_FINAL;
    list_move_tail(&c->list, &queue_info->fill_list );
    atomic_dec( &queue_info->ready_cnt );
    atomic_inc(&queue_info->fill_cnt);
    spin_unlock_irqrestore(&queue_info->list_lock, flags);

    if( (atomic_read(&queue_info->fill_cnt) > 0) && waitqueue_active(&queue_info->fill_wq) ){
        wake_up(&queue_info->fill_wq);
    }

    return 0;
}

static void _revoke_hdrl_request_write_queue( noe_drv_t *noe_drv )
{
    /*if we have noe_request_write_t , must need to free.*/
    unsigned long flags;
    noe_hdrl_data_t *c, *n;
    request_write_queue *queue = &noe_drv->request_write_queue;

    spin_lock_irqsave(&queue->list_lock, flags);
    list_for_each_entry_safe( c, n, &queue->ready_list, list ){
        list_del(&c->list);
        kmem_cache_free( queue->cache, c );
        atomic_dec( &queue->ready_cnt );
    }
    spin_unlock_irqrestore(&queue->list_lock, flags);
}

static int _wait_hdrl_queue_ready( request_write_queue *queue_info )
{
    if( wait_event_interruptible( queue_info->ready_wq, (atomic_read(&queue_info->ready_cnt)>0) ) < 0 ){
        return -ERESTARTSYS;
    }
    
    return 0;
}

static int _wait_hdrl_unit_filled( request_write_queue *queue_info, struct noe_hdrl_received_info *info )
{
    noe_write_unit_data_t *c;
    unsigned long flags;

    /* wait completion for unit is filled */
    NOE_HDRL_SVR_DBG(" [SDKTASK] wait completion for fill HDRL data.....\n");
    if( wait_event_interruptible( queue_info->fill_wq, atomic_read(&queue_info->fill_cnt)>0 ) < 0 ){
        return -ERESTARTSYS;
    }

    NOE_HDRL_SVR_DBG(" [SDKTASK] <%d> Complete for fill HDRL data.....\n", atomic_read(&queue_info->fill_cnt));

    spin_lock_irqsave(&queue_info->list_lock, flags);

    if( list_empty(&queue_info->fill_list) ){
        NOE_ERROR("MAYBE BUG!! wakeuped but list is empty\n");
        spin_unlock_irqrestore(&queue_info->list_lock, flags);
        return -ENOMEM;
    }

    c = list_first_entry( &queue_info->fill_list, noe_write_unit_data_t, list );

    memcpy( info, &(c->received_info), sizeof(struct noe_hdrl_received_info) );

    list_del(&c->list);
    atomic_dec( &queue_info->fill_cnt );
    kmem_cache_free( queue_info->cache, c );

    spin_unlock_irqrestore(&queue_info->list_lock, flags);

    return 0;
}

/************************/
/*   for NIF            */
/************************/
static NOE_UINT32 _get_uuid_nif( noe_drv_t *noe_drv, NOE_INT32 DeviceID, NOE_UCHAR *uuid, NOE_UINT32 index )
{
  NOE_INT32	d;
  NOE_pFRAME	frame;

  d = NOEhi_htos32(DeviceID);

  frame.h.HostID = current->pid;
  frame.h.SubSysID = NOE_SID_UNBIND;
  frame.h.len = sizeof(NOE_INT32);
  frame.h.cmd = NOE_DLNA_GET_UUID_NIF;
  frame.h.ctrl = NOE_MK_NIFID(index, 0);
 
  NOEki_send_noe_header_payload(noe_drv, &frame.h, &d);
  
  NOEki_wait_normal_packet(noe_drv, &frame.h, uuid);

  return(NOE_ERR_DEV2API(frame.h.cmd));
}

static NOE_UINT32 _get_net_info_nif( noe_drv_t *noe_drv, NOE_NET_INFO *netinfo, NOE_UINT32 index )
{
  NOE_pFRAME	frame;

  frame.h.HostID = current->pid;
  frame.h.SubSysID = NOE_SID_UNBIND;
  frame.h.len = 0;
  frame.h.cmd = NOE_SOC_GET_NET_INFO_NIF;
  frame.h.ctrl = NOE_MK_NIFID(index, 0);
  NOEki_send_noe_header_payload(noe_drv, &frame.h, NULL);

  NOEki_wait_normal_packet(noe_drv, &frame.h, netinfo);

  return(NOE_ERR_DEV2API(frame.h.cmd));
}

static NOE_UINT32 _set_net_info_nif( noe_drv_t *noe_drv, NOE_UINT32 num_netinfo, NOE_NET_INFO *netinfo )
{
  NOE_pFRAME	frame;
  NOE_UINT8	*p, *q;

  NOE_DP("num_netinfo = %d \n",num_netinfo);

  p = q = kmalloc(sizeof(NOE_UINT32) + sizeof(*netinfo) * num_netinfo, GFP_ATOMIC);
  NOEhi_SET_32(p, num_netinfo);

  memcpy(p, netinfo, sizeof(*netinfo) * num_netinfo);

  NOE_DP("nif_no= %d \n", netinfo->nif_no);

  frame.h.HostID = current->pid;
  frame.h.SubSysID = NOE_SID_UNBIND;
  frame.h.len = sizeof(NOE_UINT32) + sizeof(*netinfo) * num_netinfo;
  frame.h.cmd = NOE_SOC_SET_NET_INFO_NIF;
  NOEki_send_noe_header_payload(noe_drv, &frame.h, q);

  NOEki_wait_normal_packet(noe_drv, &frame.h, NULL);

  kfree(q);

  return(NOE_ERR_DEV2API(frame.h.cmd));

}

/************************/
/*   status control     */
/************************/
bool is_init_usb_status( void )
{
    return test_bit( DEF_NFE_USB_STATUS_INIT, &g_noe_usb_status );
}

/************************/
/*    NIF control       */
/************************/
#define NOE_PROCFS_NAME_CURRENT_NIF     "noe_nif"


static int _proc_nif_read( char* buf, char** start, off_t offset,
                                          int count, int* eof, void* data )
{
    size_t read_size = 0;
    noe_drv_t *noe_drv;

    noe_drv = (noe_drv_t*)data;

    read_size += snprintf( buf, 
			  sizeof(buf)+32,
                          "noe_current_nif  :%d\n",
			   (NOE_UINT32)atomic_read(&noe_drv->current_nif) );

    return read_size;

}


static int _proc_nif_write( struct file *file, const char *buffer,
                                unsigned long count, void *data )
{
    char buf[8];
    noe_drv_t *noe_drv;
    copy_from_user( (void*)buf, (void*)buffer, sizeof(buf) );

    noe_drv = (noe_drv_t*)data;

    switch(buf[0]){
    case '0': //Wired eth1
        atomic_set(&noe_drv->current_nif,0);
        break;
    case '1': //Wireless wlan0
        NOE_DP("noe_nif = 1 \n");
        atomic_set(&noe_drv->current_nif,1);
        break;
      default:
        break;
    }

    return count;
}


static void _setup_procfs_nif( noe_drv_t *noe_drv )
{
  struct proc_dir_entry *debug_nif;


  debug_nif = create_proc_entry( NOE_PROCFS_NAME_CURRENT_NIF, 0, NULL );
  if( debug_nif == NULL ){
    printk("cannot create NOE NIF procfs\n");
    return;
  }

  debug_nif->read_proc = _proc_nif_read;
  debug_nif->write_proc = _proc_nif_write;
  debug_nif->data = (void*)noe_drv;
}

static void _terminate_procfs_nif( void )
{
  remove_proc_entry( NOE_PROCFS_NAME_CURRENT_NIF, NULL );
}

/************************/
/*    debug control     */
/************************/
#define NOE_PROCFS_NAME_DEBUG_CONSOLE   "noe_console"


static int _subsystem_monitor( noe_drv_t *noe_drv, NOE_UINT32 enable )
{
    NOE_UINT32	para, _para;
    NOE_HEADER	h;

    para = noe_drv->current_console_para;

    printk("noe_debug enable = %d \n", enable);
    printk("noe_debug current_console_para = %d \n", noe_drv->current_console_para);
  
    switch(enable){
    case 0:
      printk("noe_debug case0 \n");
      para = NOEhi_htos32(NOE_MONITOR_DISABLE);
	break;

    case 1:
      printk("noe_debug case1 \n");
      _para = NOE_MONITOR_ENABLE;
      para =  NOEhi_htos32( _para | noe_drv->current_console_para );
        break;

    case 2:
      printk("noe_debug case2 \n");
      _para = NOE_MONITOR_TCPIP + NOE_MONITOR_ENABLE ;
      para =  NOEhi_htos32( _para | noe_drv->current_console_para );
        break;

    case 3:
      printk("noe_debug case3 \n");
      _para = NOE_MONITOR_DTCP + NOE_MONITOR_ENABLE;
      para =  NOEhi_htos32( _para | noe_drv->current_console_para );
        break;

    case 4:
      printk("noe_debug case4 \n");
      _para = NOE_MONITOR_HTTP + NOE_MONITOR_ENABLE;
      para =  NOEhi_htos32( _para | noe_drv->current_console_para );
        break;

    case 5:
      printk("noe_debug case5 \n");
      _para = NOE_MONITOR_UPNP + NOE_MONITOR_ENABLE;
      para =  NOEhi_htos32( _para | noe_drv->current_console_para );
        break;

    default:
      printk("noe_debug default \n");
      para = NOEhi_htos32(NOE_MONITOR_DISABLE);
	break;      
    }

    noe_drv->current_console_para = para;

    h.SubSysID = NOE_SID_UNBIND;
    h.HostID = current->pid;
    h.len = sizeof(NOE_UINT32);
    h.cmd = NOE_CONTROL_MONITOR;

    NOEki_send_noe_header_payload( noe_drv, &h, &para );
    printk("noe_debug monitor_param=%d \n",para);

    NOEki_wait_normal_packet( noe_drv, &h, NULL );

    if(h.cmd != NOE_ACK) {
      printk("cannot setup subsystem monitor\n");
      return -1;
    }

    return 0;
}

static int _proc_console_write( struct file *file, const char *buffer,
                                unsigned long count, void *data )
{
    char buf[8];
    noe_drv_t *noe_drv;


    copy_from_user( (void*)buf, (void*)buffer, sizeof(buf) );
    printk("noe_debug buf = %d \n", buf[0]);

    noe_drv = (noe_drv_t*)data;

    switch(buf[0]){
      case '0':
          g_noe_debug_console = 0;
          _subsystem_monitor( noe_drv, 0 );
        break;
      case '1':
          g_noe_debug_console = 1;
          _subsystem_monitor( noe_drv, 1 );
	break;
      case '2':
          g_noe_debug_console = 2;
          _subsystem_monitor( noe_drv, 2 );
	break;
      case '3':
          g_noe_debug_console = 3;
          _subsystem_monitor( noe_drv, 3 );
	break;
      case '4':
          g_noe_debug_console = 4;
          _subsystem_monitor( noe_drv, 4 );
	break;
      case '5':
          g_noe_debug_console = 5;
          _subsystem_monitor( noe_drv, 5 );
        break;
      default:
        break;
    }

    return count;
}

static void _setup_procfs( noe_drv_t *noe_drv )
{
  struct proc_dir_entry *debug_console;

  debug_console = create_proc_entry( NOE_PROCFS_NAME_DEBUG_CONSOLE, 0, NULL );
  if( debug_console == NULL ){
    printk("cannot create NOE console procfs\n");
    return;
  }

  debug_console->write_proc = _proc_console_write;
  debug_console->data = (void*)noe_drv;
}

static void _terminate_procfs( noe_drv_t *noe_drv )
{
  remove_proc_entry( NOE_PROCFS_NAME_DEBUG_CONSOLE, NULL );
}

static void _start_network_monitor_thread( noe_drv_t *noe_drv )
{
  /* sorry... not implemented */
}

static void _stop_network_monitor_thread( noe_drv_t *noe_drv )
{
  /* sorry... not implemented */
}

static void _noe_data_dump( noe_drv_t *noe_drv, int pos, NOE_UINT32 HostID )
{
  /* sorry... not implemented */
}

/* table of devices that work with this driver */
static struct usb_device_id noe_device_table[] = {
	{
	USB_DEVICE(USB_VENDOR_ID, NOE_PRODUCT_ID),
	.driver_info =	(unsigned long)1,
	},
	{}
};
MODULE_DEVICE_TABLE(usb, noe_device_table);

static struct usb_driver noe_driver = {
	.name =		driver_name,
	.probe =	noe_probe,
	.disconnect =	noe_disconnect,
	.id_table =	noe_device_table,
	.suspend =	noe_suspend,
	.resume =	noe_resume
};

static int __init noe_drv_init(void)
{
        int i;

        /* added by sony */
        for( i = 0; i < NOE_DEVICE_TYPE_MAX; i++ ){
            INIT_LIST_HEAD(&g_netdevinfo_list[i]);
        }
        init_waitqueue_head(&g_rmmod_ready_wq);

	return usb_register(&noe_driver);
}

static void __exit noe_drv_exit(void)
{
        clear_bit( DEF_NFE_USB_STATUS_INIT, &g_noe_usb_status ); /* added by sony */

        _wait_rmmod_ready(); /* added by sony */

	usb_deregister(&noe_driver);
}

module_init(noe_drv_init);
module_exit(noe_drv_exit);

MODULE_LICENSE("GPL");
